# packages
import warnings
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import List, Union, Tuple, Optional, Dict
from enum import Enum

# qis
import qis.utils.struct_ops as sop
import qis.plots.utils as put
from qis.plots.utils import LegendStats, TrendLine, LastLabel
from qis.perfstats.desc_table import compute_desc_table, DescTableType


def plot_time_series(df: Union[pd.Series, pd.DataFrame],
                     linestyles: List[str] = None,
                     linewidth: float = 1.0,
                     x_date_freq: Optional[str] = 'YE',
                     date_format: str = '%d-%b-%y',
                     legend_title: str = None,
                     legend_loc: Optional[Union[str, bool]] = 'upper left',
                     last_label: LastLabel = LastLabel.NONE,
                     sort_by_value_stretch_factor: float = 1.0,
                     trend_line: Optional[TrendLine] = TrendLine.NONE,
                     trend_line_colors: List[str] = None,
                     legend_stats: LegendStats = LegendStats.AVG_LAST,
                     desc_table_type: DescTableType = DescTableType.NONE,
                     legend_labels: List[str] = None,
                     indices_for_shaded_areas: Dict[str, Tuple[int, int]] = None,
                     xlabel: str = None,
                     ylabel: str = None,
                     var_format: Optional[str] = '{:,.2f}',
                     markers: List[str] = False,
                     title: Union[str, bool] = None,
                     fontsize: int = 10,
                     markersize: int = None,
                     colors: List[str] = None,
                     x_limits: Tuple[Optional[float], Optional[float]] = None,
                     y_limits: Tuple[Optional[float], Optional[float]] = None,
                     is_log: bool = False,
                     ax: plt.Subplot = None,
                     **kwargs
                     ) -> Optional[plt.Figure]:

    data1 = df.copy()
    if isinstance(data1, pd.DataFrame):
        pass
    elif isinstance(data1, pd.Series):
        data1 = data1.to_frame()
    else:
        raise TypeError(f"unsuported data type {type(data1)}")
    columns = data1.columns

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if colors is None:
        colors = put.get_n_colors(n=len(columns), **kwargs)

    sns.lineplot(data=data1, palette=colors, dashes=False, markers=markers,
                 markersize=markersize,
                 linewidth=linewidth, ax=ax)

    # add tredlines
    if trend_line_colors is None:
        trend_line_colors = colors
    if trend_line in [TrendLine.ZERO_SHADOWS, TrendLine.ABOVE_ZERO_SHADOWS]:
        for column, color in zip(columns, trend_line_colors):
            x0 = data1.index[0]
            x1 = data1.index[-1]
            y0 = 0
            y1 = 0
            x = [x0, x1]
            y = [y0, y1]
            ax.plot(x, y,
                    color='black',
                    linestyle='-',
                    transform=ax.transData,
                    lw=linewidth)

            x_ = (data1.index - data1.index[0]).days
            slope = (y1 - y0) / (x_[-1] - x_[0])
            y_line = [slope * x + y0 for x in x_]
            y = data1[column]
            if trend_line == TrendLine.ZERO_SHADOWS:
                ax.fill_between(data1.index, y, y_line, where=y_line >= y,
                            facecolor=color, interpolate=True, alpha=0.2, lw=linewidth)
            else:
                ax.fill_between(data1.index, y, y_line, where=y_line <= y,
                                facecolor=color, interpolate=True, alpha=0.2, lw=linewidth)

    elif trend_line in [TrendLine.AVERAGE, TrendLine.AVERAGE_SHADOWS]:
        for column, color in zip(columns, trend_line_colors):
            data2 = data1[column].dropna()  # exclude nans from showing the average lines
            if data2.empty:  # skip this columns from ere
                continue
            else:
                average = np.nanmean(data2.to_numpy())

            x0 = data2.index[0]
            x1 = data2.index[-1]
            y0 = average
            y1 = average
            x = [x0, x1]
            y = [y0, y1]

            ax.plot(x, y,
                    color=color,
                    linestyle=':',
                    transform=ax.transData,
                    linewidth=linewidth)

            if trend_line == TrendLine.AVERAGE_SHADOWS:
                x_ = (data2.index - data2.index[0]).days
                slope = (y1 - y0) / (x_[-1] - x_[0])
                y_line = [slope * x + y0 for x in x_]
                y = data2.to_numpy()
                ax.fill_between(data2.index, y, y_line, where=y_line >= y,
                                facecolor=color, interpolate=True, alpha=0.2, lw=linewidth)

    elif trend_line in [TrendLine.TREND_LINE, TrendLine.TREND_LINE_SHADOWS]:
        for column, color in zip(columns, trend_line_colors):
            y = data1[column].dropna()
            if not y.empty:
                x0 = y.first_valid_index()
                x1 = y.index[-1]
                y0 = y[x0]
                y1 = y.iloc[-1]
                x = [x0, x1]
                y = [y0, y1]
                ax.plot(x, y,
                        color=color,
                        linestyle='--',
                        transform=ax.transData,
                        linewidth=linewidth)

                if trend_line == TrendLine.TREND_LINE_SHADOWS:
                    x_ = (data1.index - data1.index[0]).days
                    slope = (y1 - y0) / (x_[-1] - x_[0])
                    y_line = [slope * x + y0 for x in x_]
                    y = data1[column]
                    ax.fill_between(data1.index, y, y_line, where=y_line >= y,
                                    facecolor=color, interpolate=True, alpha=0.2, lw=linewidth)

    # add last labels
    if last_label in [LastLabel.AVERAGE_VALUE, LastLabel.AVERAGE_VALUE_SORTED]:
        average_dict = {}
        for column, color in zip(columns, colors):
            data2 = data1[column].dropna()  # exclude nans from showing the average lines

            if data2.empty:  # skip this columns from ere
                continue
            else:
                average = np.nanmean(data2.to_numpy())

            if var_format is not None:
                average_str = var_format.format(average)
            else:
                average_str = average
            y1 = average
            variable_label = f"{column}, average={average_str}"
            average_dict.update({y1: [variable_label, color]})

        x1 = data1.index[-1]
        ymin, ymax = ax.get_ylim()
        mid = sort_by_value_stretch_factor * (ymax - ymin)
        if last_label == LastLabel.AVERAGE_VALUE_SORTED:
            pivot_dict = sorted(average_dict)
            locs = np.linspace(pivot_dict[0], sort_by_value_stretch_factor*mid, len(pivot_dict), endpoint=True)
        else:
            pivot_dict = average_dict
            locs = [dict for dict in pivot_dict]

        for key, loc in zip(pivot_dict, locs):
            ax.annotate(average_dict[key][0],
                        xy=(x1, key), xytext=(x1, loc),
                        fontsize=fontsize, weight='normal', color=average_dict[key][1],
                        textcoords='data', ha='left', va='bottom',
                        bbox={'boxstyle': 'round,pad=0.5', 'fc': average_dict[key][1], 'alpha': 0.1},
                        arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))

    elif last_label in [LastLabel.LAST_VALUE, LastLabel.LAST_VALUE_SORTED]:
        last_dict = {}
        for column, color in zip(columns, colors):
            y = data1[column].dropna()
            if last_label == LastLabel.LAST_VALUE:
                if var_format is not None:
                    if len(y.index) > 0:
                        variable_str = var_format.format(y.iloc[-1])
                    else:
                        variable_str = 'nan'
                else:
                    variable_str = y.iloc[-1]
                variable_label = f"{column}, last = {variable_str}"

                if len(y.index) > 0:
                    y1 = y.iloc[-1]
                else:
                    y1 = np.nan
                last_dict.update({y1: [variable_label, color]})

        # plot dicts sorted by last value
        x1 = data1.index[-1]
        ymin, ymax = ax.get_ylim()
        mid = sort_by_value_stretch_factor * (ymax - ymin)

        if last_label == LastLabel.LAST_VALUE_SORTED:
            pivot_dict = sorted(last_dict)
            locs = np.linspace(pivot_dict[0], sort_by_value_stretch_factor*mid, len(pivot_dict), endpoint=True)
        else:
            pivot_dict = last_dict
            locs = [dict for dict in pivot_dict]

        for key, loc in zip(pivot_dict,locs):
            ax.annotate(last_dict[key][0],
                        xy=(x1, key), xytext=(x1, loc),
                        fontsize=fontsize, weight ='normal', color = last_dict[key][1],
                        textcoords='data', ha='left', va='bottom',
                        bbox={'boxstyle': 'round,pad=0.5', 'fc': last_dict[key][1], 'alpha': 0.1},
                        arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))

    if indices_for_shaded_areas is not None:
        if len(data1.columns) == 1:
            warnings.warn(f"cannot do indices_for_shaded_areas = {indices_for_shaded_areas} for dataframe with one columns ")
        else:
            for col, indxs in indices_for_shaded_areas.items():
                y0 = data1.iloc[:, indxs[0]]
                y1 = data1.iloc[:, indxs[1]]
                ax.fill_between(data1.index, y0, y1, where=y0 >= y1,
                                facecolor=col, alpha=0.2,
                                interpolate=True)

    if legend_loc is not None:
        if legend_labels is None:
            if desc_table_type != DescTableType.NONE:  # use get_legend_with_stats_table
                stats_table = compute_desc_table(df=data1,
                                                 desc_table_type=desc_table_type,
                                                 var_format=var_format)
                put.set_legend_with_stats_table(stats_table=stats_table,
                                                ax=ax,
                                                colors=colors,
                                                legend_loc=legend_loc,
                                                fontsize=fontsize,
                                                **kwargs)
            else:  # generate legend_labels
                legend_labels = put.get_legend_lines(data=data1, legend_stats=legend_stats, var_format=var_format)

        if legend_labels is not None:
            put.set_legend(ax=ax,
                           labels=legend_labels,
                           colors=colors,
                           legend_loc=legend_loc,
                           legend_title=legend_title,
                           fontsize=fontsize,
                           **kwargs)

    else:
        ax.legend().set_visible(False)

    if linestyles is not None:
        put.set_linestyles(ax=ax, linestyles=linestyles)

    if x_date_freq is not None and isinstance(data1.index, pd.DatetimeIndex):
        put.set_date_on_axis(data=data1,
                             ax=ax,
                             x_date_freq=x_date_freq,
                             date_format=date_format,
                             fontsize=fontsize,
                             **kwargs)

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)

    if x_limits is not None:
        put.set_x_limits(ax=ax, x_limits=x_limits)

    if var_format is not None:
        put.set_ax_ticks_format(ax=ax, fontsize=fontsize, xvar_format=None, yvar_format=var_format, **kwargs)

    if isinstance(data1.index, pd.DatetimeIndex):
        put.set_ax_tick_labels(ax=ax, skip_y_axis=True, fontsize=fontsize, **kwargs)

    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)
    put.set_spines(ax=ax, **kwargs)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    if is_log:
        ax.set_yscale('log')

    return fig


def plot_time_series_2ax(df1: Union[pd.Series, pd.DataFrame],
                         df2: Union[pd.Series, pd.DataFrame],
                         legend_loc: Optional[str] = 'upper left',
                         legend_stats: LegendStats = LegendStats.NONE,
                         legend_stats2: LegendStats = LegendStats.NONE,
                         title: Optional[str] = None,
                         var_format: str = '{:,.0f}',
                         var_format_yax2: str = '{:,.0f}',
                         ylabel1: str = None,
                         ylabel2: str = None,
                         x_date_freq: Union[str, None] = 'YE',
                         legend_labels: List[str] = None,
                         linestyles: List[str] = None,
                         linestyles_ax2: List[str] = None,
                         y_limits: Tuple[Optional[float], Optional[float]] = None,
                         y_limits_ax2: Tuple[Optional[float], Optional[float]] = None,
                         trend_line1: put.TrendLine = put.TrendLine.NONE,
                         trend_line2: put.TrendLine = put.TrendLine.NONE,
                         yvar_major_ticks1: np.ndarray = None,
                         yvar_major_ticks2: np.ndarray = None,
                         colors: List[str] = None,
                         fontsize: int = 10,
                         x_rotation: int = 90,
                         is_logs: Tuple[bool, bool] = (False, False),
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if isinstance(df1, pd.Series):
        df1 = df1.to_frame()
    if isinstance(df2, pd.Series):
        df2 = df2.to_frame()

    ncols1 = len(df1.columns)
    ncols2 = len(df2.columns)
    if colors is None:
        colors = put.get_n_colors(n=ncols1 + ncols2, **kwargs)

    ax_twin = ax.twinx()
    plot_time_series(df=df1,
                     legend_loc=None,
                     colors=colors[:ncols1],
                     var_format=None,
                     linestyles=linestyles,
                     trend_line=trend_line1,
                     x_date_freq=x_date_freq,
                     is_log=is_logs[0],
                     y_limits=y_limits,
                     ylabel=ylabel1,
                     fontsize=fontsize,
                     ax=ax,
                     **kwargs)

    plot_time_series(df=df2,
                     legend_loc=None,
                     colors=colors[ncols1:],
                     var_format=None,
                     linestyles=linestyles_ax2,
                     trend_line=trend_line2,
                     x_date_freq=x_date_freq,
                     is_log=is_logs[1],
                     y_limits=y_limits_ax2,
                     ylabel=ylabel2,
                     fontsize=fontsize,
                     ax=ax_twin,
                     **kwargs)

    ax.tick_params(axis='x', which='both', bottom=False)

    put.set_ax_ticks_format(ax=ax, fontsize=fontsize, xvar_format=None, yvar_format=var_format, set_ticks=False,
                            yvar_major_ticks=yvar_major_ticks1, x_rotation=x_rotation, **kwargs)
    put.set_ax_ticks_format(ax=ax_twin, fontsize=fontsize, xvar_format=None, yvar_format=var_format_yax2, set_ticks=False,
                            yvar_major_ticks=yvar_major_ticks2, x_rotation=x_rotation, **kwargs)

    if legend_loc is not None:
        if legend_labels is None:
            df1.columns = [f"{x} (left)" for x in df1.columns]
            df2.columns = [f"{x} (right)" for x in df2.columns]
            legend_labels1 = put.get_legend_lines(data=df1,
                                                  legend_stats=legend_stats,
                                                  var_format=var_format)
            legend_labels2 = put.get_legend_lines(data=df2,
                                                  legend_stats=legend_stats2,
                                                  var_format=var_format_yax2)
            legend_labels = sop.to_flat_list(legend_labels1 + legend_labels2)
        put.set_legend(ax=ax,
                       labels=legend_labels,
                       colors=colors,
                       fontsize=fontsize,
                       legend_loc=legend_loc,
                       **kwargs)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize)

    return fig


class LocalTests(Enum):
    PRICES = 1
    PRICES_2AX = 2


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if local_test == LocalTests.PRICES:
        fig, axs = plt.subplots(2, 1, figsize=(8, 6), tight_layout=True)
        global_kwargs = {'fontsize': 8,
                         'linewidth': 0.5,
                         'weight': 'normal',
                         'markersize': 1}
        plot_time_series(df=prices,
                         legend_stats=put.LegendStats.AVG_LAST,
                         last_label=LastLabel.AVERAGE_VALUE_SORTED,
                         trend_line=TrendLine.AVERAGE_SHADOWS,
                         ax=axs[0],
                         **global_kwargs)
        plot_time_series(df=prices,
                         legend_stats=put.LegendStats.AVG_LAST,
                         last_label=LastLabel.LAST_VALUE,
                         trend_line=TrendLine.AVERAGE_SHADOWS,
                         ax=axs[1],
                         **global_kwargs)

    elif local_test == LocalTests.PRICES_2AX:
        fig, ax = plt.subplots(1, 1, figsize=(4, 3), tight_layout=True)
        global_kwargs = {'fontsize': 6,
                         'linewidth': 0.5,
                         'weight': 'normal',
                         'markersize': 1}

        plot_time_series_2ax(df1=prices.iloc[:, -1],
                             df2=prices.iloc[:, :-1],
                             legend_stats=put.LegendStats.AVG_LAST,
                             var_format_yax2='{:.0f}',
                             ax=ax,
                             **global_kwargs)

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.PRICES)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Tuple, List
import qis as qis


def get_test_data(n_days: int = 1000) -> Tuple[pd.Series, pd.Series]:
    """
    generate randpm nav and volume
    """
    index = pd.date_range(start='1Jan2020', periods=n_days, freq='D')
    returns = pd.Series(np.random.normal(0.0, 0.2/np.sqrt(260), n_days), index=index, name='nav')
    nav = qis.returns_to_nav(returns=returns)
    volume = returns.abs().cumsum().rename('volume')
    return nav, volume


def generate_figure_2x(num_assets: int = 5, n_days: int = 1000) -> List[plt.Figure]:
    """
    plot figure for each asst
    """
    figs = []
    for n in np.arange(num_assets):
        nav, volume = get_test_data(n_days=n_days)
        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 1, figsize=(16, 12), tight_layout=True)
            figs.append(fig)
            title = f"asset num {n+1}"
            qis.plot_time_series_2ax(df1=nav, df2=volume,
                                     var_format='{:,.2f}',
                                     var_format_yax2='{:,.2f}',
                                     title=title, ax=ax)
    return figs


figs = generate_figure_2x(num_assets=5, n_days=1000)
qis.save_figs_to_pdf(figs=figs, file_name='test_price_report', local_path=qis.get_output_path())



import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
from bbg_fetch import fetch_field_timeseries_per_tickers

# tickers = {'CF1 Index': 'CAC', 'GX1 Index': 'DAX', 'IB1 Index': 'IBEX', 'ST1 Index': 'MIB', 'EO1 Index': 'AEX'}
tickers = {'CF1 Index': 'CAC', 'GX1 Index': 'DAX', 'IB1 Index': 'IBEX', 'ST1 Index': 'MIB'}
start_date = pd.Timestamp('1Jan1990')

futures_prices = fetch_field_timeseries_per_tickers(tickers=tickers, freq='B', field='PX_LAST',
                                                    start_date=start_date).ffill()
# use 3m rolling volumes
volumes = fetch_field_timeseries_per_tickers(tickers=tickers, freq='B', field='VOLUME',
                                             start_date=start_date).ffill().rolling(63).mean()
price_volumes = volumes.multiply(futures_prices).dropna(how='all')

# create volume-weighted portfolios
futures_weights = qis.df_to_weight_allocation_sum1(df=price_volumes.asfreq('QE').ffill())

# plot futures price data
kwargs = dict(x_date_freq='YE')
with sns.axes_style("darkgrid"):
    fig, axs = plt.subplots(2, 2, figsize=(10, 10), tight_layout=True)
    qis.plot_prices_with_dd(prices=futures_prices, axs=axs[:, 0], **kwargs)
    qis.plot_time_series(df=volumes, title='Contact volumes', ax=axs[0, 1], **kwargs)
    qis.plot_time_series(df=price_volumes, title='Value volumes', ax=axs[1, 1], **kwargs)


# add long cash SPTR
sptr_prices = fetch_field_timeseries_per_tickers(tickers={'SPTR Index': 'SPTR'}, freq='B', field='PX_LAST',
                                                 start_date=start_date).ffill()
long_sptr_short_futures_weights = pd.concat([pd.Series(1.0, index=futures_weights.index, name='SPTR'),
                                             -1.0*futures_weights], axis=1)
print(long_sptr_short_futures_weights)

prices = pd.concat([sptr_prices, futures_prices], axis=1)

# backtest total return portfolio
time_period = qis.TimePeriod('31Dec1990', '27Sep2024')
portfolio_data = qis.backtest_model_portfolio(prices=time_period.locate(prices),
                                              weights=time_period.locate(long_sptr_short_futures_weights),
                                              rebalancing_costs=0.0005,
                                              weight_implementation_lag=1,
                                              ticker='Long SPTR / Short EU futures')

# create factcheet
figs = qis.generate_strategy_factsheet(portfolio_data=portfolio_data,
                                       benchmark_prices=sptr_prices.iloc[:, 0].rename('SPTR long').to_frame(),
                                       add_current_position_var_risk_sheet=True,
                                       time_period=time_period,
                                       **qis.fetch_default_report_kwargs(time_period=time_period))
qis.save_figs_to_pdf(figs=figs,
                     file_name=f"long_sptr_short_eu_strategy_factsheet",
                     local_path=qis.local_path.get_output_path())

plt.show()



"""
core for computing performance
"""
# packages
import warnings
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from enum import Enum
from typing import Union, Dict, Optional

import qis.utils.annualisation
# qis
import qis.utils.dates as da
import qis.utils.df_freq as dff
import qis.utils.np_ops as npo
import qis.utils.df_ops as dfo
from qis.perfstats.config import PerfStat, ReturnTypes, PerfParams
from qis.utils.dates import CALENDAR_DAYS_PER_YEAR_SHARPE


def compute_num_days(prices: Union[pd.DataFrame, pd.Series]) -> int:
    if prices.index[0] > prices.index[-1]:
        raise ValueError(f"inconsistent dates t0={prices.index[0]} t1={prices.index[-1]}")
    return np.maximum((prices.index[-1] - prices.index[0]).days, 1)


def compute_num_years(prices: Union[pd.DataFrame, pd.Series],
                      days_per_year: float = CALENDAR_DAYS_PER_YEAR_SHARPE
                      ) -> float:
    return compute_num_days(prices=prices) / days_per_year


def to_returns(prices: Union[pd.Series, pd.DataFrame],
               is_log_returns: bool = False,
               return_type: ReturnTypes = ReturnTypes.RELATIVE,
               freq: Optional[str] = None,
               include_start_date: bool = False,
               include_end_date: bool = False,
               ffill_nans: bool = True,
               drop_first: bool = False,
               is_first_zero: bool = False,
               **kwargs
               ) -> Union[pd.Series, pd.DataFrame]:
    """
    compute periodic returns using return_type
    """
    prices = prices_at_freq(prices=prices, freq=freq,
                            include_start_date=include_start_date,
                            include_end_date=include_end_date,
                            ffill_nans=ffill_nans)

    # returns are computed for non nans and positive prices
    prices_np = prices.to_numpy()
    ind_good = np.logical_and(np.isnan(prices_np) == False, np.greater(prices_np, 0.0))

    if return_type == ReturnTypes.LOG or is_log_returns:
        returns = np.log(prices).diff(1)

    elif return_type == ReturnTypes.RELATIVE:
        returns = np.divide(prices, prices.shift(1)).add(-1.0)

    elif return_type == ReturnTypes.DIFFERENCE:
        returns = prices - prices.shift(1)

    elif return_type == ReturnTypes.LEVEL:
        returns = prices

    elif return_type == ReturnTypes.LEVEL0:
        returns = prices.shift(1)
    else:
        raise NotImplementedError (f"{return_type}")

    returns = returns.where(ind_good, other=np.nan)

    if is_first_zero:
        if isinstance(returns, pd.DataFrame):
            returns.iloc[0, :] = 0
        else:
            returns.iloc[0] = 0
    elif drop_first:
        if isinstance(returns, pd.DataFrame):
            returns = returns.iloc[1:, :]
        else:
            returns = returns.iloc[1:]

    return returns


def to_total_returns(prices: Union[pd.Series, pd.DataFrame]) -> pd.Series:
    total_returns = compute_total_return(prices=prices)
    if isinstance(prices, pd.DataFrame):
        total_returns = pd.Series(total_returns, index=prices.columns)
    elif isinstance(prices, pd.Series):
        total_returns = pd.Series(total_returns, name=prices.name)
    else:
        raise NotImplementedError(f"{type(prices)}")

    return total_returns


def compute_total_return(prices: Union[pd.DataFrame, pd.Series]) -> Union[np.ndarray, float]:
    """
    np.ndarray for pd.DataFrame
    float for pd.Series
    """
    if len(prices.index) == 1:
        if isinstance(prices, pd.DataFrame):
            return np.full(len(prices.columns), fill_value=np.nan)
        elif isinstance(prices, pd.Series):
            return np.nan
        else:
            raise TypeError(f"unsuported type={type(prices)}")

    if isinstance(prices, pd.DataFrame):
        price_0 = prices.iloc[0, :].to_numpy()
        if np.any(np.isnan(price_0)):
            price_0 = dfo.get_first_nonnan_values(df=prices)
            print(f"detected nan price for prices = {prices.iloc[0, np.isnan(price_0)]},"
                  f" using first non nan price = {price_0} for {prices.columns}")

        price_end = prices.iloc[-1, :].to_numpy()

    elif isinstance(prices, pd.Series):
        price_0 = prices.iloc[0]
        if np.isnan(price_0):
            price_0 = dfo.get_first_nonnan_values(df=prices)
            print(f"detected nan price for prices = {prices.iloc[0]},"
                  f" using first non nan price = {price_0} for {prices.name}")

        price_end = prices.iloc[-1]

    else:
        raise TypeError(f"unsuported type={type(prices)}")

    num_years = compute_num_years(prices=prices)
    if num_years > 0.0:
        total_return = price_end / price_0 - 1.0
    else:
        print(prices)
        print(f"total return has incosistent dates t0={prices.index[0]} and t1={prices.index[-1]}")
        total_return = np.full(len(prices.columns), fill_value=np.nan) if isinstance(prices, pd.DataFrame) else np.nan
    return total_return


def compute_pa_return(prices: Union[pd.DataFrame, pd.Series],
                      annualize_less_1y: bool = False
                      ) -> Union[np.ndarray, float]:
    """
    np.ndarray for pd.DataFrame
    float for pd.Series
    """
    total_return = compute_total_return(prices=prices)
    num_years = compute_num_years(prices=prices)

    if num_years > 0.0:
        ratio = total_return + 1.0
        ratio = np.where(np.greater(ratio, 0.0), ratio, np.nan)
        if num_years > 1.0:
            compounded_return_pa = np.power(ratio, 1.0 / num_years, where=np.isfinite(ratio)) - 1
        else:
            if annualize_less_1y:  # annualize to 1y, not compound
                compounded_return_pa = total_return / num_years
            else:  # use ytd return
                compounded_return_pa = ratio - 1.0

    else:
        n = len(prices.columns) if isinstance(prices, pd.DataFrame) else 1
        compounded_return_pa = np.zeros_like(n)

    return compounded_return_pa


def compute_returns_dict(prices: Union[pd.DataFrame, pd.Series],
                         perf_params: PerfParams = None,
                         annualize_less_1y: bool = False
                         ) -> Dict[str, np.ndarray]:
    """
    compute returns for one asset
    """
    if not isinstance(prices, pd.Series) and not isinstance(prices, pd.DataFrame):
        raise ValueError(f"not supperted type={type(prices)}")

    if prices.empty:
        warnings.warn(f"in compute_returns_dict(): {prices} is all nans", stacklevel=2)
        if isinstance(prices, pd.Series):
            n = 1
        else:
            n = len(prices.columns)
        if n == 1:
            return_dict = {PerfStat.TOTAL_RETURN.to_str(): np.nan,
                           PerfStat.PA_RETURN.to_str(): np.nan,
                           PerfStat.AN_LOG_RETURN.to_str(): np.nan,
                           PerfStat.AN_LOG_RETURN_EXCESS.to_str(): np.nan,
                           PerfStat.APR.to_str(): np.nan,
                           PerfStat.NUM_YEARS.to_str(): np.nan}
        else:
            return_dict = {PerfStat.TOTAL_RETURN.to_str(): np.full(n, fill_value=np.nan),
                           PerfStat.PA_RETURN.to_str(): np.full(n, fill_value=np.nan),
                           PerfStat.AN_LOG_RETURN.to_str(): np.full(n, fill_value=np.nan),
                           PerfStat.AN_LOG_RETURN_EXCESS.to_str(): np.full(n, fill_value=np.nan),
                           PerfStat.APR.to_str(): np.full(n, fill_value=np.nan),
                           PerfStat.NUM_YEARS.to_str(): np.full(n, fill_value=np.nan)}
        return return_dict

    if perf_params is None:  # only needed for EXCESS returns use defaults
        perf_params = PerfParams()

    compounded_return_pa = compute_pa_return(prices=prices, annualize_less_1y=annualize_less_1y)
    total_return = compute_total_return(prices=prices)
    num_years = compute_num_years(prices=prices)
    num_days = compute_num_days(prices=prices)

    if perf_params.rates_data is not None:
        excess_return_pa = compute_pa_excess_compounded_returns(returns=to_returns(prices,
                                                                        return_type=ReturnTypes.RELATIVE,
                                                                        is_first_zero=True),
                                                     rates_data=perf_params.rates_data,
                                                     first_date=prices.index[0],
                                                     annualize_less_1y=annualize_less_1y)
    else:
        excess_return_pa = compounded_return_pa

    if isinstance(prices, pd.DataFrame):
        start_value = prices.iloc[0, :].to_numpy()
        end_value = prices.iloc[-1, :].to_numpy()
    else:
        start_value = prices.iloc[0]
        end_value = prices.iloc[-1]

    # make dictionary output
    return_dict = {PerfStat.TOTAL_RETURN.to_str(): total_return,
                   PerfStat.PA_RETURN.to_str(): compounded_return_pa,
                   PerfStat.PA_EXCESS_RETURN.to_str(): excess_return_pa,
                   PerfStat.AN_LOG_RETURN.to_str(): np.log(1.0 + compounded_return_pa, where=np.greater(compounded_return_pa, -1.0)),
                   PerfStat.AN_LOG_RETURN_EXCESS.to_str(): np.log(1.0 + excess_return_pa, where=np.greater(excess_return_pa, -1.0)),
                   PerfStat.AVG_AN_RETURN.to_str(): np.divide(total_return, num_years),
                   PerfStat.APR.to_str(): CALENDAR_DAYS_PER_YEAR_SHARPE*total_return/num_days if num_days > 0 else CALENDAR_DAYS_PER_YEAR_SHARPE*total_return,
                   PerfStat.NAV1.to_str(): (1.0+total_return),
                   PerfStat.NUM_YEARS.to_str(): num_years,
                   PerfStat.START_DATE.to_str(): prices.index[0],
                   PerfStat.END_DATE.to_str(): prices.index[-1],
                   PerfStat.START_PRICE.to_str(): start_value,
                   PerfStat.END_PRICE.to_str(): end_value}

    return return_dict


def compute_excess_return_navs(prices: Union[pd.Series, pd.DataFrame],
                               rates_data: pd.Series,
                               first_date: pd.Timestamp = None
                               ) -> Union[pd.Series, pd.DataFrame]:
    # get returns and subtract average rate between dt times:
    returns = to_returns(prices=prices, is_first_zero=True)
    excess_returns = compute_excess_returns(returns=returns, rates_data=rates_data)
    navs = returns_to_nav(returns=excess_returns, first_date=first_date)
    return navs


def compute_excess_returns(returns: Union[pd.Series, pd.DataFrame],
                           rates_data: pd.Series
                           ) -> Union[pd.Series, pd.DataFrame]:
    # get returns and subtract average rate between dt times:
    rates_dt = dfo.multiply_df_by_dt(df=rates_data, dates=returns.index, lag=None)
    returns0 = returns.copy()
    if isinstance(returns, pd.Series):
        returns0 = returns0.to_frame(name=returns.name)
    excess_returns = returns0.subtract(rates_dt.to_numpy(), axis=0)
    if isinstance(returns, pd.Series):
        excess_returns = excess_returns.iloc[:, 0]
    return excess_returns


def compute_pa_excess_compounded_returns(returns: Union[pd.Series, pd.DataFrame],
                                         rates_data: pd.Series,
                                         first_date: pd.Timestamp = None,
                                         annualize_less_1y: bool = False
                                         ) -> Union[np.ndarray, float]:
    excess_returns = compute_excess_returns(returns=returns, rates_data=rates_data)
    prices = returns_to_nav(returns=excess_returns, first_date=first_date)
    compounded_return_pa = compute_pa_return(prices=prices, annualize_less_1y=annualize_less_1y)
    if isinstance(compounded_return_pa, np.ndarray):
        compounded_return_pa = compounded_return_pa[0]
    return compounded_return_pa


def estimate_vol(sampled_returns: Union[pd.DataFrame, pd.Series, np.ndarray]) -> np.ndarray:
    if isinstance(sampled_returns, np.ndarray):
        n = sampled_returns.shape[0]
    else:
        n = len(sampled_returns.index)
    if n >= 20:  # adjust for mean for small sample
        vol = np.nanstd(sampled_returns, axis=0, ddof=1)
    else: # take sum
        vol = np.sqrt(np.nanmean(np.power(sampled_returns, 2), axis=0))
    return vol


def compute_sampled_vols(prices: Union[pd.DataFrame, pd.Series],
                         freq_vol: str = 'ME',
                         freq_return: Optional[str] = None,
                         include_start_date: bool = False,
                         include_end_date: bool = False,
                         ) -> Union[pd.DataFrame, pd.Series]:

    """
    compute close to close vol based on return samples
    """
    sampled_returns = to_returns(prices=prices, freq=freq_return,
                                 include_start_date=include_start_date, include_end_date=include_end_date)
    # sampled_returns_at_vol_freq = sampled_returns.resample(freq_vol)
    sampled_returns_at_vol_freq = da.split_df_by_freq(df=sampled_returns,
                                                      freq=freq_vol,
                                                      overlap_frequency=None,
                                                      include_start_date=include_start_date,
                                                      include_end_date=include_end_date)

    vol_samples = {}
    for key, df in sampled_returns_at_vol_freq.items():
        vol_samples[key] = estimate_vol(sampled_returns=df.to_numpy())

    if isinstance(prices, pd.Series):
        vols = pd.DataFrame.from_dict(vol_samples, orient='index').iloc[:, 0].rename(prices.name)
    else:
        vols = pd.DataFrame.from_dict(vol_samples, orient='index', columns=prices.columns)
    vols = vols.multiply(np.sqrt(qis.utils.annualisation.infer_an_from_data(sampled_returns)))

    return vols


def portfolio_navs_to_additive(grouped_nav: pd.DataFrame,
                               portfolio_name: str
                               ) -> pd.DataFrame:
    """
    given pandas with portfolio total and asset navs adjust asset navs
    """
    portfolio_nav = grouped_nav[portfolio_name]
    ac_nav_adj = adjust_navs_to_portfolio_pa(portfolio_nav=portfolio_nav,
                                             asset_prices=grouped_nav.drop(columns=[portfolio_name]))
    grouped_nav = pd.concat([portfolio_nav, ac_nav_adj], axis=1)
    return grouped_nav


def adjust_navs_to_portfolio_pa(portfolio_nav: pd.Series,
                                asset_prices: pd.DataFrame
                                ) -> pd.DataFrame:
    """
    adjust navs of portfolio navs so that the total pa return on navs are additive to portfolio total
    adjustment to match terminal value backpropagated: more stable
    """
    portfolio_pa = compute_pa_return(prices=portfolio_nav)
    assets_pa = compute_pa_return(prices=asset_prices)
    n = len(asset_prices.columns)
    asset_prices_adj = asset_prices.copy()
    t = (portfolio_nav.index - portfolio_nav.index[0]).days.to_numpy() / CALENDAR_DAYS_PER_YEAR_SHARPE
    c_m = ((portfolio_pa / n + 1.0) / (np.nanmean(assets_pa) + 1.0)) ** t
    ratio = npo.np_array_to_df_columns(a=c_m, ncols=n)
    asset_prices_adj = asset_prices_adj.multiply(ratio)
    asset_prices_adj = asset_prices_adj[asset_prices.columns]
    return asset_prices_adj


def compute_net_return(gross_return: pd.Series,
                       man_fee: float = 0.01,
                       perf_fee: float = 0.2,
                       perf_fee_frequency: str = 'YE'
                       ) -> pd.Series:
    """
    compute fee adjusted returns from gross returns
    """
    perf_fee_cristalization_schedule = da.generate_dates_schedule(time_period=da.TimePeriod(gross_return.index[0], gross_return.index[-1]),
                                                                  freq=perf_fee_frequency)

    perf_cris_dates = np.isin(element=gross_return.index,
                              test_elements=perf_fee_cristalization_schedule,
                              assume_unique=True)

    nav_data = pd.DataFrame(data=0.0,
                            index=gross_return.index,
                            columns=['Net Return', 'NAV', 'GAV', 'PF', 'HWM', 'CPF'])
    nav_data.insert(loc=0, column='gross return', value=gross_return.to_numpy())
    nav_data = nav_data.copy()

    nav_data.loc[nav_data.index[0], 'GAV'] = 100.0
    nav_data.loc[nav_data.index[0], 'NAV'] = 100.0
    nav_data.loc[nav_data.index[0], 'HWM'] = 100.0

    for date, last_date, perf_cris_date in zip(nav_data.index[1:], nav_data.index[0:], perf_cris_dates[1:]):

        man_fee_dt = man_fee * (date-last_date).days/365.0

        nav_data.loc[date, 'GAV'] = (1.0+nav_data.loc[date, 'gross return']-man_fee_dt)*nav_data.loc[last_date, 'GAV']

        nav_data.loc[date, 'PF'] = perf_fee*np.maximum(nav_data.loc[date, 'GAV']-nav_data.loc[last_date, 'HWM'], 0.0)
        nav_data.loc[date, 'NAV'] = nav_data.loc[date, 'GAV']-nav_data.loc[date, 'PF']
        nav_data.loc[date, 'HWM'] = nav_data.loc[last_date, 'HWM']

        if perf_cris_date:
            nav_data.loc[date, 'CPF'] = nav_data.loc[date, 'PF']
            nav_data.loc[date, 'HWM'] = np.maximum(nav_data.loc[date, 'NAV'], nav_data.loc[last_date, 'HWM'] )
            nav_data.loc[date, 'GAV'] = nav_data.loc[date, 'GAV'] - nav_data.loc[date, 'CPF']

    net_return = nav_data['NAV'] / nav_data['NAV'].shift(1)-1.0
    net_return = net_return.copy()
    net_return.iloc[0] = 0.0
    net_return = net_return.rename(gross_return.name)

    return net_return


def get_net_navs(navs: Union[pd.Series, pd.DataFrame],
                 man_fee: float = 0.01,
                 perf_fee: float = 0.2,
                 perf_fee_frequency: str = 'YE'
                 ) -> Union[pd.Series, pd.DataFrame]:

    gross_returns = navs.pct_change()
    net_returns = []
    if isinstance(navs, pd.Series):
        net_returns = compute_net_return(gross_return=gross_returns,
                                         man_fee=man_fee,
                                         perf_fee=perf_fee,
                                         perf_fee_frequency=perf_fee_frequency)
    else:
        for column in gross_returns.columns:
            net = compute_net_return(gross_return=gross_returns[column],
                                     man_fee=man_fee,
                                     perf_fee=perf_fee,
                                     perf_fee_frequency=perf_fee_frequency)
            net_returns.append(net)
        net_returns = pd.concat(net_returns, axis=1)
    net_nav = returns_to_nav(returns=net_returns)
    return net_nav


def returns_to_nav(returns: Union[np.ndarray, pd.Series, pd.DataFrame],
                   init_period: Optional[int] = 1,
                   terminal_value: np.ndarray = None,
                   init_value: Union[np.ndarray, float] = None,
                   first_date: pd.Timestamp = None,
                   freq: Optional[str] = None,
                   constant_trade_level: bool = False,
                   ffill_between_nans: bool = True   # for nan returns fill between nans
                   ) -> Union[pd.Series, pd.DataFrame]:
    """
    instrument returns to nav
    init_period of one by default will exlude the first day of returns for compunded nav
    """
    if init_period is not None and isinstance(returns, np.ndarray) is False:
        returns = to_zero_first_nonnan_returns(returns=returns, init_period=init_period)
    elif first_date is not None:
        # the nav=100 will start from the first date if given
        # if returns[0] is not zero then first nav will be discounted from 100
        if isinstance(returns, pd.DataFrame):
            returns.loc[:first_date, :] = 0.0
        elif isinstance(returns, pd.Series):
            returns.loc[:first_date] = 0.0

    if constant_trade_level:
        strategy_nav = returns.cumsum(skipna=True, axis=0).add(1.0)
    else:
        if isinstance(returns, np.ndarray):
            strategy_nav = np.cumprod(1.0+returns, axis=0)
        else:
            strategy_nav = returns.add(1.0).cumprod(skipna=True)

    if terminal_value is not None:
        terminal_value_last = dfo.get_last_nonnan_values(strategy_nav)
        strategy_nav = strategy_nav*(terminal_value/terminal_value_last)
    elif init_value is not None:
        initial_value_first = dfo.get_first_nonnan_values(df=strategy_nav)
        strategy_nav = strategy_nav*(init_value / initial_value_first)

    if freq is not None and isinstance(returns, np.ndarray) is False:  # when it is important to have fixed frequency ffill prices
        strategy_nav = strategy_nav.asfreq(freq, method='ffill').ffill()

    if ffill_between_nans and isinstance(returns, np.ndarray) is False:
        strategy_nav = df_price_ffill_between_nans(prices=strategy_nav, method='ffill')

    return strategy_nav


def prices_to_scaled_nav(prices: Union[pd.Series, pd.DataFrame], scale=0.5):
    """
    rescale price returns by scale
    """
    returns = scale * to_returns(prices=prices, is_log_returns=False, is_first_zero=True)
    navs = returns_to_nav(returns=returns)
    return navs


def prices_at_freq(prices: Union[pd.Series, pd.DataFrame],
                   freq: Optional[str] = None,
                   include_start_date: bool = False,
                   include_end_date: bool = False,
                   ffill_nans: bool = True,
                   fill_na_method: Optional[str] = 'ffill'
                   ) -> Union[pd.Series, pd.DataFrame]:
    """
    get prices at freq
    """
    if freq is not None:
        if ffill_nans:
            fill_na_method = 'ffill'
        else:
            fill_na_method = None
        prices = dff.df_asfreq(df=prices,
                               freq=freq,
                               include_start_date=include_start_date,
                               include_end_date=include_end_date,
                               fill_na_method=fill_na_method)
    else:
        if fill_na_method is not None:
            if fill_na_method == 'ffill':
                prices = prices.ffill()
            elif fill_na_method == 'bfill':
                prices = prices.bfill()
            else:
                raise NotImplementedError(f"fill_na_method={fill_na_method}")

    return prices


def log_returns_to_nav(log_returns: Union[np.ndarray, pd.Series, pd.DataFrame],
                       init_period: Optional[int] = None,
                       terminal_value: np.ndarray = None,
                       init_value: Union[np.ndarray, float] = None
                       ) -> Union[pd.Series, pd.DataFrame]:
    """
    instrument log_return returns to nav
    init_period of one by default will exlude the first day of returns for compounded nav
    """
    if init_period is not None and isinstance(log_returns, np.ndarray) is False:
        log_returns = to_zero_first_nonnan_returns(returns=log_returns, init_period=init_period)

    strategy_nav = np.exp(log_returns.cumsum(axis=0, skipna=True))

    if terminal_value is not None:
        strategy_nav = strategy_nav.multiply(terminal_value/dfo.get_last_nonnan_values(df=strategy_nav))
    elif init_value is not None:
        strategy_nav = strategy_nav.divide(dfo.get_first_nonnan_values(df=strategy_nav) / init_value)

    return strategy_nav


def long_short_to_relative_nav(long_price: pd.Series, short_price: pd.Series) -> pd.Series:
    """
    performance of strategy long_price - short_price
    """
    returns = to_returns(pd.concat([long_price, short_price], axis=1).ffill(), is_first_zero=True)
    relative_returns = np.subtract(returns[long_price.name], returns[short_price.name])
    relative_nav = returns_to_nav(returns=relative_returns, init_period=1)
    return relative_nav


def to_portfolio_returns(weights: pd.DataFrame,
                         returns: pd.DataFrame,
                         portfolio_name: str = 'portfolios'
                         ) -> pd.Series:
    """
    instrument returns to portfolio, equivalent to with
    portfolio_returns = returns.multiply(weights.shift(1)).sum(axis=1)
    but with handling of nan instead of zero when using .sum(axis=1)
    """
    weights_1 = weights.shift(1)
    portfolio_pnl = returns.multiply(weights_1).to_numpy()
    pnl = np.nansum(portfolio_pnl, axis=1)
    is_all_nan = np.all(np.isnan(portfolio_pnl), axis=1)
    portfolio_pnl = np.where(is_all_nan, np.nan, pnl)
    portfolio_returns = pd.Series(data=portfolio_pnl,
                                  index=returns.index,
                                  name=portfolio_name)
    return portfolio_returns


def portfolio_returns_to_nav(returns: pd.DataFrame,
                             init_period: Optional[int] = 1,
                             init_value: Union[np.ndarray, float] = None,
                             freq: Optional[str] = None
                             ) -> Union[pd.Series, pd.DataFrame]:
    """
    instrument returns to nav
    """

    agg_pnl = pd.Series(np.nansum(returns.to_numpy(), axis=1), index=returns.index)
    nav = returns_to_nav(returns=agg_pnl, init_period=init_period, init_value=init_value, freq=freq)
    return nav


def compute_grouped_nav(returns: pd.DataFrame,
                        init_period: Union[int, None] = None,
                        init_value: float = 1.0,
                        freq: Optional[str] = None
                        ) -> pd.Series:
    """
    use implementation returns_to_nav()
    """
    grouped_nav = portfolio_returns_to_nav(returns=returns,
                                           init_period=init_period,
                                           init_value=init_value,
                                           freq=freq)
    return grouped_nav


def to_zero_first_nonnan_returns(returns: Union[pd.Series, pd.DataFrame],
                                 init_period: Union[int, None] = 1
                                 ) -> Union[pd.Series, pd.DataFrame]:
    """
    replace first nonnan return with zero
    """
    if init_period is not None and not isinstance(init_period, int):
        raise ValueError(f"init_period must be integer")

    if init_period is not None:
        if init_period == 1:
            first_before_nonnan_index = dfo.get_nonnan_index(df=returns, position='first')
            first_date = returns.index[0]
            if isinstance(returns, pd.Series):
                if first_before_nonnan_index >= first_date:
                    returns[first_before_nonnan_index] = 0.0
            else:
                for first_before_nonnan_index_, column in zip(first_before_nonnan_index, returns.columns):
                    if first_before_nonnan_index_ >= first_date:
                        returns.loc[first_before_nonnan_index_, column] = 0.0
        else:
            warnings.warn(f"in returns_to_nav init_period={init_period} is not supported")

    return returns


def get_excess_returns_nav(prices: Union[pd.DataFrame, pd.Series],
                           funding_rate: pd.Series,
                           freq: str = 'B'
                           ) -> Union[pd.DataFrame, pd.Series]:

    if not isinstance(funding_rate, pd.Series):
        raise ValueError(f"funding_rate must be series")

    nav_returns = to_returns(prices=prices, freq=freq)

    funding_rate_dt = dfo.multiply_df_by_dt(df=funding_rate, dates=nav_returns.index, lag=1)

    if isinstance(prices, pd.DataFrame):
        ncols = len(prices.columns)
        excess_returns = nav_returns.subtract(npo.np_array_to_df_columns(a=funding_rate_dt.to_numpy(), ncols=ncols),
                                              axis=1)
    else:
        data = nav_returns.to_numpy()-funding_rate_dt.to_numpy()
        excess_returns = pd.Series(data, index=nav_returns.index, name=nav_returns.name)

    terminal_value = dfo.get_last_nonnan_values(prices)
    excess_nav = returns_to_nav(returns=excess_returns,
                                    terminal_value=terminal_value,
                                    init_period=1)
    return excess_nav


def df_price_ffill_between_nans(prices: Union[pd.Series, pd.DataFrame],
                                method: Optional[str] = 'ffill'
                                ) -> Union[pd.Series, pd.DataFrame]:
    """
    fill prices between nans for date0 = first_nonnan_dte to date1 = last_nonnan_date
    """
    is_series_out = False
    if isinstance(prices, pd.Series):
        is_series_out = True
        prices = prices.to_frame()

    first_date = dfo.get_nonnan_index(df=prices, position='first')
    last_date = dfo.get_nonnan_index(df=prices, position='last')
    good_parts = []
    for idx, column in enumerate(prices.columns):
        good_price = prices.loc[first_date[idx]:last_date[idx], column]
        if method is not None:
            good_price = good_price.ffill()
        good_parts.append(good_price)
    bfilled_data = pd.concat(good_parts, axis=1)
    if bfilled_data.index[0] > prices.index[0]:
        bfilled_data = bfilled_data.reindex(index=prices.index)
    if is_series_out:
        bfilled_data = bfilled_data.iloc[:, 0]
    return bfilled_data


class LocalTests(Enum):
    TO_ZERO_NONNAN = 1
    VOL_SAMPLE = 2
    ADJUST_PORTFOLIO_PA_RETURNS = 3
    NET_RETURN = 4
    ROLLING_RETURNS = 5


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    import qis.plots.time_series as pts
    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if local_test == LocalTests.TO_ZERO_NONNAN:
        np.random.seed(2)  # freeze seed
        dates = pd.date_range(start='31Dec2020', end='07Jan2021', freq='B')
        n = 3
        returns = pd.DataFrame(data=np.random.normal(0.0, 0.01, (len(dates), n)),
                               index=dates,
                               columns=['x' + str(m + 1) for m in range(n)])

        returns.iloc[:, 0] = np.nan
        returns.iloc[:2, 1] = np.nan
        returns.iloc[:1, 2] = np.nan
        returns.iloc[3, 2] = np.nan

        print(f"returns:\n{returns}")

        returns1 = to_zero_first_nonnan_returns(returns=returns)
        print(f"zero_first_non_nan_returns=\n{returns1}")

        navs = returns_to_nav(returns=returns)
        print(f"navs with init_period = 1:\n{navs}")
        navs = returns_to_nav(returns=returns, init_period=None)
        print(f"navs with init_period = None:\n{navs}")

    elif local_test == LocalTests.VOL_SAMPLE:
        vols = compute_sampled_vols(prices=prices,
                                    freq_return='B',
                                    freq_vol='ME')
        print(vols)

    elif local_test == LocalTests.ADJUST_PORTFOLIO_PA_RETURNS:
        returns = prices.pct_change()

        portfolio_price = returns_to_nav(returns=returns.sum(1)).rename('portfolio')

        asset_prices_adj = adjust_navs_to_portfolio_pa(portfolio_nav=portfolio_price,
                                                       asset_prices=prices)

        asset_prices_adj.columns = [x + ' adjusted' for x in asset_prices_adj.columns]

        plot_data = pd.concat([prices.divide(prices.iloc[0, :], axis=1),
                               asset_prices_adj.divide(asset_prices_adj.iloc[0, :], axis=1),
                               portfolio_price], axis=1)
        pts.plot_time_series(df=plot_data,
                             var_format='{:.2f}',
                             title='Original vs Adjusted NAVs')
        print(asset_prices_adj)

    elif local_test == LocalTests.NET_RETURN:
        nav = prices['SPY'].dropna()
        print(nav)
        net_navs = get_net_navs(navs=nav)
        print(net_navs)

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.NET_RETURN)



"""
line plot
"""
# packages
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import List, Union, Tuple, Optional, Dict
from enum import Enum
import qis.plots.utils as put


def plot_line(df: Union[pd.Series, pd.DataFrame],
              x: str = None,
              y: str = None,
              hue: str = None,
              linestyles: List[str] = None,
              linestyle: str = '-',
              linewidth: float = 1.0,
              legend_title: str = None,
              legend_loc: Optional[Union[str, bool]] = 'upper left',
              legend_stats: put.LegendStats = put.LegendStats.NONE,
              legend_labels: List[str] = None,
              x_labels: List[str] = None,
              xlabel: str = None,
              ylabel: str = None,
              title: str = None,
              xvar_format: Optional[str] = None,  # '{:,.2f}', # unless data is numerical
              yvar_format: Optional[str] = '{:,.2f}',
              markers: Union[str, List[str]] = False,
              fontsize: int = 10,
              colors: List[str] = None,
              x_limits: Tuple[Optional[float], Optional[float]] = None,
              y_limits: Tuple[Optional[float], Optional[float]] = None,
              is_log: bool = False,
              ax: plt.Subplot = None,
              **kwargs
              ) -> Optional[plt.Figure]:
    """
    line plot wrapper for sns, lineplot
    plot df using index
    x: str = None: x column
    y: str = None: y column
    hue: str = None: hue
    """
    if isinstance(df, pd.DataFrame):
        pass
    elif isinstance(df, pd.Series):
        df = df.to_frame()
    else:
        raise TypeError(f"unsuported data type {type(df)}")

    if x is not None:
        if y is None:
            raise ValueError(f"y column must be given with x column")
        else:
            if hue is not None:
                df = df[[x, y, hue]]
            else:
                df = df[[x, y]]

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if colors is None:
        if hue is None:
            colors = put.get_n_colors(n=len(df.columns), **kwargs)
        else:
            colors = put.get_n_colors(n=len(df[hue].unique()), **kwargs)

    sns.lineplot(data=df, x=x, y=y, hue=hue,
                 palette=colors, dashes=False, markers=markers, linestyle=linestyle, linewidth=linewidth,
                 style=hue,
                 ax=ax)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    if legend_loc is not None:
        if legend_labels is None:
            if hue is None:
                legend_labels = put.get_legend_lines(data=df, legend_stats=legend_stats, var_format=yvar_format)
            else:
                h, legend_labels = ax.get_legend_handles_labels()
        put.set_legend(ax=ax,
                       labels=legend_labels,
                       colors=colors,
                       legend_loc=legend_loc,
                       legend_title=legend_title,
                       markers=markers,
                       fontsize=fontsize,
                       **kwargs)

    else:
        ax.legend().set_visible(False)

    if linestyles is not None:
        put.set_linestyles(ax=ax, linestyles=linestyles)

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)
    if x_limits is not None:
        put.set_x_limits(ax=ax, x_limits=x_limits)

    put.set_ax_ticks_format(ax=ax, fontsize=fontsize, xvar_format=xvar_format, yvar_format=yvar_format)

    if x_labels is None:
        put.set_ax_ticks_format(ax=ax, fontsize=fontsize, **kwargs)

    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)
    put.set_spines(ax=ax, **kwargs)

    if is_log:
        ax.set_yscale('log')

    return fig


def plot_lines_list(xy_datas: Dict[str, pd.DataFrame],
                    data_labels: List[List[str]],
                    colors: List[str] = None,
                    markers: List[str] = None,
                    is_fill_annotations: bool = False,
                    xvar_format: str = '{:.1f}',
                    yvar_format: str = '{:.1f}',
                    xlabel: str = None,
                    ylabel: str = None,
                    title: str = None,
                    fontsize: int = 10,
                    x_limits: Tuple[Union[float, None], Union[float, None]] = None,
                    y_limits: Tuple[Union[float, None], Union[float, None]] = None,
                    legend_loc: str = 'upper left',
                    ax: plt.Subplot = None,
                    **kwargs
                    ) -> plt.Figure:
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if colors is None:
        colors = put.get_n_colors(n=len(xy_datas.keys()))

    for idx, (key, xy) in enumerate(xy_datas.items()):
        color = colors[idx]
        labels = data_labels[idx]

        if markers is not None:
            marker = markers[idx]
        else:
            marker = 'None'
        sns.lineplot(x=xy.columns[0], y=xy.columns[1], data=xy, marker=marker, color=color, ax=ax)

        for label, x, y in zip(labels, xy.iloc[:, 0].to_numpy(), xy.iloc[:, 1].to_numpy()):
            if label != '':
                x_offset = 1
                y_offset = 1
                if is_fill_annotations:
                    ax.annotate(
                        label,
                        xy=(x, y), xytext=(x_offset, y_offset),
                        textcoords='offset points', ha='right', va='bottom',
                        bbox=dict(boxstyle='round, pad=0.5', fc='blue', alpha=0.75),
                        arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'),
                        fontsize=fontsize)
                else:
                    ax.annotate(
                        label,
                        xy=(x, y), xytext=(x_offset, y_offset),
                        textcoords='offset points', ha='right', va='bottom',
                        fontsize=fontsize)

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)

    if x_limits is not None:
        put.set_x_limits(ax=ax, x_limits=x_limits)

    put.set_ax_ticks_format(ax=ax, fontsize=fontsize, xvar_format=xvar_format, yvar_format=yvar_format)
    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize)

    if legend_loc is not None:
        legend_labels = list(xy_datas.keys())
        put.set_legend(ax=ax,
                       labels=legend_labels,
                       colors=colors,
                       legend_loc=legend_loc,
                       markers=markers,
                       fontsize=fontsize,
                       **kwargs)
    else:
        ax.legend().set_visible(False)

    put.set_spines(ax=ax, **kwargs)

    return fig


class LocalTests(Enum):
    LINEPLOT = 1
    MOVE_DATA = 2


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    if local_test == LocalTests.LINEPLOT:
        x = np.linspace(0, 14, 100)
        y = np.sin(x)
        data = pd.Series(y, index=x, name='data')
        fig, axs = plt.subplots(2, 1, figsize=(8, 6))
        global_kwargs = dict(fontsize=12, linewidth=2.0, weight='normal', markersize=2)
        plot_line(df=data, legend_stats=put.LegendStats.AVG_LAST, ax=axs[0], **global_kwargs)
        plot_line(df=data, legend_stats=put.LegendStats.AVG_LAST,
                  linestyle='dotted',
                  ax=axs[1], **global_kwargs)

    elif local_test == LocalTests.MOVE_DATA:
        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(2, 1, figsize=(12, 6))
            global_kwargs = dict(fontsize=12, linewidth=2.0, weight='normal', markersize=2)
            index = ['06-07JUN22', '10-17JUN22', '17-24JUN22', '24JUN-01JUL22', '26JUN-30SEP22', '30SEP-30DEC22']
            data = [0.7644, 0.7602, 0.7306, 0.7524, 0.8192, 0.8204]
            df = pd.DataFrame(data, index=index, columns=['Expected Move % annualized'])
            print(df)
            markers = put.get_n_markers(n=1)
            plot_line(df=df,
                      title='ATM forward volatilities Implied from BTC MOVE contracts on 06-Jun-2022',
                      legend_stats=put.LegendStats.NONE,
                      yvar_format='{:.0%}',
                      xvar_format=None,
                      markers=markers,
                      ax=axs[0],
                      **global_kwargs)

            plot_line(df=df,
                      title='ATM forward volatilities Implied from BTC MOVE contracts on 06-Jun-2022',
                      legend_stats=put.LegendStats.NONE,
                      yvar_format='{:.0%}',
                      xvar_format=None,
                      linewidth=0.,
                      markers=['s'] * len(df.columns),
                      ax=axs[1])

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.MOVE_DATA)



"""
run simulation of short exposure strategies
follow the discussion in
https://twitter.com/ArturSepp/status/1614551490093359104
to run the code, install qis package (Quantitative Investment Strategies):
pip install qis
"""

# imports
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import yfinance as yf
import qis

# load dprice data for given ticker
ticker = 'SPY'
price = yf.download(ticker, start="2003-12-31", end=None, ignore_tz=True, auto_adjust=True)['Close'].rename(ticker)
price = price.loc['2016':]
price_np = price.to_numpy()

# specify position
long_or_short = -1.0
constant_notional = 1.0

# start backtest
# this will track constant notional strategy
constant_notional_units, constant_notional_cum_nav = np.zeros_like(price_np), np.zeros_like(price_np)
constant_notional_units[0], constant_notional_cum_nav[0] = long_or_short * constant_notional / price_np[0], constant_notional

# this will track constant nav exposure strategy which is equivalent to long or short leverage ETF
short_etf_units, short_etf_cum_nav = np.zeros_like(price_np), np.zeros_like(price_np)
short_etf_units[0], short_etf_cum_nav[0] = long_or_short * constant_notional / price_np[0], constant_notional

for idx, (price0, price1) in enumerate(zip(price_np[:-1], price_np[1:])):
    # time_t = idx+1
    constant_notional_cum_nav[idx+1] = constant_notional_cum_nav[idx] + constant_notional_units[idx] * (price1-price0)
    # exposure in unit is constant for the next step and computed using eod price
    constant_notional_units[idx+1] = long_or_short*constant_notional / price1

    short_etf_cum_nav[idx+1] = short_etf_cum_nav[idx] + short_etf_units[idx] * (price1-price0)
    # etf is rebalanced proportionally to the current eod nav
    short_etf_units[idx+1] = long_or_short*short_etf_cum_nav[idx+1]/price1

# store
constant_notional_cum_nav = pd.Series(constant_notional_cum_nav, index=price.index, name='constant_notional_cum_nav')
short_etf_cum_nav = pd.Series(short_etf_cum_nav, index=price.index, name='short_etf_cum_nav')

# add simple byu and hold starts
buy_and_hold_units = np.ones_like(price_np) * constant_notional / price_np[0]
buy_and_hold_cum_nav = pd.Series(constant_notional + buy_and_hold_units*(price_np-price_np[0]), index=price.index, name='buy_and_hold_cum_nav')

sell_and_hold_units = - np.ones_like(price_np) * constant_notional / price_np[0]
sell_and_hold_cum_nav = pd.Series(constant_notional + sell_and_hold_units*(price_np-price_np[0]), index=price.index, name='sell_and_hold_cum_nav')

# prices
prices = pd.concat([buy_and_hold_cum_nav, sell_and_hold_cum_nav, constant_notional_cum_nav, short_etf_cum_nav], axis=1)

# portfolio units
buy_and_hold_units = pd.Series(buy_and_hold_units, index=price.index, name='buy_and_hold_units')
sell_and_hold_units = pd.Series(sell_and_hold_units, index=price.index, name='sell_and_hold_units')
constant_notional_units = pd.Series(constant_notional_units, index=price.index, name='constant_notional_units')
short_etf_units = pd.Series(short_etf_units, index=price.index, name='short_etf_units')
portfolio_units = pd.concat([buy_and_hold_units, sell_and_hold_units, constant_notional_units, short_etf_units], axis=1)


with sns.axes_style("darkgrid"):
    fig, axs = plt.subplots(2, 1, figsize=(9, 7), tight_layout=True)

    # plot performance
    qis.plot_prices_with_dd(prices=prices,
                            perf_stats_labels=qis.PerfStatsLabels.TOTAL.value,
                            title=f"Realized performance of strategies with short exposure to {ticker}",
                            axs=axs)

    # plot units
    fig, ax = plt.subplots(1, 1, figsize=(9, 7), tight_layout=True)
    qis.plot_time_series(df=portfolio_units,
                         title='Portfolio Units',
                         ax=ax)

    fig, ax = plt.subplots(1, 1, figsize=(9, 7), tight_layout=True)
    qis.plot_returns_scatter(prices=prices,
                             benchmark=prices.columns[0],
                             freq='W-MON',
                             return_type=qis.ReturnTypes.DIFFERENCE,  # cannot use log and relative returns
                             title='Scatterplot of weekly P&L relative to buy-and-hold',
                             ylabel='Daily P&L of short strategies',
                             xlabel='Daily P&L of buy-and-hold',
                             var_format='{:,.0%}',
                             ax=ax)

plt.show()



"""
imlementation of basic linear factor model
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from dataclasses import dataclass
from typing import Union, Dict, Literal, Tuple, Optional
# qis
from qis.utils.dates import TimePeriod, find_upto_date_from_datetime_index
from qis.utils.struct_ops import merge_lists_unique
import qis.models.linear.ewm as ewm
import qis.perfstats.returns as ret
import qis.plots.time_series as pts
from qis.portfolio.risk.contributions import calculate_marginal_active_risk

@dataclass
class LinearModel:
    """A linear factor model for analyzing asset returns using multiple factors.

    This class implements a linear factor model of the form y = loadings^T @ x + alpha,
    where factors (x) explain asset returns (y) through time-varying factor loadings.
    The model supports factor attribution analysis, alpha calculation, model diagnostics,
    and performance evaluation.

    The linear relationship assumes:
    - y_t = β_t^T * x_t + α_t
    - Where β_t are the factor loadings (betas) at time t
    - x_t are the factor returns at time t
    - α_t is the unexplained return (alpha) at time t

    Attributes:
        x (pd.DataFrame or pd.Series): Factor data of shape (T, N) where T is time periods
            and N is number of factors. Each column represents a different factor's
            time series returns.
        y (pd.DataFrame or pd.Series): Asset data of shape (T, M) where T is time periods
            and M is number of assets. Each column represents a different asset's
            time series returns.
        loadings (Dict[str, pd.DataFrame], optional): Estimated factor loadings for each
            factor. Keys are factor names (matching x.columns), values are DataFrames
            with shape (T, M) containing time-varying loadings for each asset.

    Methods:
        Core Analysis:
            - get_factor_loadings(): Retrieve loadings for specific factors
            - get_asset_factor_betas(): Get factor exposures for specific assets
            - get_asset_factor_attribution(): Calculate factor attribution for returns
            - get_factor_alpha(): Calculate unexplained returns (alpha) and explained returns

        Portfolio Analysis:
            - compute_agg_factor_exposures(): Aggregate factor exposures using portfolio weights

        Model Diagnostics:
            - get_model_ewm_r2(): Calculate exponentially weighted R-squared
            - get_model_residuals_corrs(): Analyze residual correlations

        Visualization:
            - plot_factor_loadings(): Plot factor loading time series
            - print(): Debug output of model components

    Example:
        ```python
        # Create factor and asset return data
        factors = pd.DataFrame({
            'Market': market_returns,
            'Size': size_returns,
            'Value': value_returns
        }, index=dates)

        assets = pd.DataFrame({
            'Stock_A': stock_a_returns,
            'Stock_B': stock_b_returns
        }, index=dates)

        # Assume loadings have been estimated elsewhere
        loadings = {
            'Market': market_betas_df,  # Shape: (T, 2) for 2 assets
            'Size': size_betas_df,
            'Value': value_betas_df
        }

        # Create model
        model = LinearModel(x=factors, y=assets, loadings=loadings)

        # Analyze factor attribution for Stock_A
        attribution = model.get_asset_factor_attribution(asset='Stock_A')

        # Calculate model alpha (unexplained returns)
        alpha, explained = model.get_factor_alpha(lag=1)

        # Check model fit
        r2 = model.get_model_ewm_r2(span=52)
        ```
        See example implementation with EwmLinearModel
    Notes:
        - Factor loadings are assumed to be pre-estimated (e.g., via rolling regression)
        - Time series must be aligned with matching indices
        - The model supports both in-sample (lag=0) and out-of-sample (lag=1) analysis
        - All DataFrames should have datetime indices for proper time series handling
        - Missing values in loadings or returns may affect calculations

    Raises:
        AssertionError: If loadings keys don't match factor column names in x
    """

    x: Union[pd.DataFrame, pd.Series] # t, x_n factors
    y: Union[pd.DataFrame, pd.Series] # t, y_m factors
    loadings: Dict[str, pd.DataFrame] = None  # estimated factor loadings
    x_covars: Optional[Dict[pd.Timestamp, pd.DataFrame]] = None  # variances of x factors
    residual_vars: Optional[pd.DataFrame] = None  # residual variance of y

    def __post_init__(self):

        if isinstance(self.x, pd.Series):
            self.x = self.x.to_frame()

        if self.y is not None and isinstance(self.y, pd.Series):
            self.y = self.y.to_frame()

        if self.loadings is not None:
            str_factors = list(self.loadings.keys())
            assert str_factors == self.x.columns.to_list()

    def print(self):
        """Print model components for debugging."""
        print(f"x:\n{self.x}")
        print(f"y:\n{self.y}")
        for factor, loading in self.loadings.items():
            print(f"{factor}:\n{loading}")

    def get_factor1_loadings(self) -> pd.DataFrame:
        """Get loadings for the first factor."""
        return self.loadings[list(self.loadings.keys())[0]]

    def get_factor_loadings(self, factor: str) -> pd.DataFrame:
        """Get loadings for a specific factor."""
        return self.loadings[factor]

    def get_loadings_at_date(self, date: pd.Timestamp) -> pd.DataFrame:
        if self.loadings is None:
            raise ValueError
        betas = {}
        for factor, df in self.loadings.items():
            last_update_date = find_upto_date_from_datetime_index(index=df.index, date=date)
            betas[factor] = df.loc[last_update_date, :]
        betas = pd.DataFrame.from_dict(betas, orient='index')  # index by factor
        return betas

    def compute_agg_factor_exposures(self,
                                     weights: pd.DataFrame
                                     ) -> pd.DataFrame:
        """Compute aggregate factor exposures by multiplying loadings with weights."""
        factor_exposures = {}
        for factor, loading in self.loadings.items():
            weights1 = weights.reindex(index=loading.index).ffill()
            factor_exposures[factor] = loading.multiply(weights1).sum(axis=1)
        factor_exposures = pd.DataFrame.from_dict(factor_exposures)
        return factor_exposures

    def get_asset_factor_betas(self,
                               time_period: TimePeriod = None,
                               asset: str = None
                               ) -> pd.DataFrame:
        """Get asset exposures to factors over specified time period."""
        if asset is None:
            asset = self.y.columns[0]
        exps = {}
        for factor, factor_exp in self.loadings.items():
            exps[factor] = factor_exp[asset]
        exps = pd.DataFrame.from_dict(exps)
        if time_period is not None:
            exps = time_period.locate(exps)
        return exps

    def get_asset_factor_attribution(self, asset: str = None, add_total: bool = True) -> pd.DataFrame:
        """Calculate factor attribution for an asset's returns."""
        factor_betas = self.get_asset_factor_betas(asset=asset)
        exposures = self.x
        attribution = exposures.multiply(factor_betas.shift(1))
        if add_total:
            total = attribution.sum(1).rename('Total')
            attribution = pd.concat([total, attribution], axis=1)
        return attribution

    def get_factor_alpha(self,
                         x: Optional[pd.DataFrame] = None,
                         y: Optional[pd.DataFrame] = None,
                         lag: Literal[0, 1] = 1,
                         span: Optional[int] = None
                         ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Calculate factor alpha and explained returns.

        Args:
            lag: 1 for step-ahead forecast, 0 for in-sample.

        Returns:
            Tuple of (factor_alpha, explained_returns).
        """
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        assert x.index.equals(y.index)

        explained_returns = pd.DataFrame(0.0, index=y.index, columns=y.columns)
        for factor in x.columns:
            factor_betas = self.loadings[factor].reindex(index=x.index).ffill()
            explained_return = (factor_betas.shift(lag)).multiply(x[factor].to_numpy(), axis=0)
            explained_returns = explained_returns.add(explained_return)
        factor_alpha = y.subtract(explained_returns)
        if span is not None:
            factor_alpha = ewm.compute_ewm(data=factor_alpha, span=span)
        return factor_alpha, explained_returns

    def get_model_ewm_r2(self, span: int = 52, lag: Literal[0, 1] = 0) -> pd.DataFrame:
        """Calculate exponentially weighted R-squared using residuals."""
        residuals, explained_returns = self.get_factor_alpha(lag=lag)
        ewm_residuals_2 = ewm.compute_ewm(data=np.square(residuals), span=span)
        y_demean = self.y  #- ewm.compute_ewm(data=self.y, span=span)
        ewm_variance = ewm.compute_ewm(data=np.square(y_demean), span=span)
        r_2 = 1.0 - ewm_residuals_2.divide(ewm_variance)
        r_2 = r_2.clip(0.0, 1.0)
        return r_2

    def get_model_residuals_corrs(self, span: int = 52) -> Tuple[pd.DataFrame, pd.Series]:
        """Calculate EWM correlation matrix of residuals and average correlation."""
        residuals, explained_returns = self.get_factor_alpha(lag=0)
        corr = ewm.compute_ewm_covar(residuals.to_numpy(), span=span, is_corr=True)
        avg_corr = pd.Series(0.5*np.nanmean(corr - np.eye(corr.shape[0]), axis=1), index=self.y.columns)
        corr_pd = pd.DataFrame(corr, index=self.y.columns, columns=self.y.columns)
        return corr_pd, avg_corr

    def compute_factor_risk_contribution(self, weights: pd.DataFrame,
                                         factor_var_name: str = 'Systematic',
                                         idiosyncratic_var_name: str = 'Idiosyncratic'
                                         ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """Compute factor risk contributions and their ratios for portfolio weights over time.
        Calculates the contribution of each factor to portfolio risk using factor loadings,
        covariance matrices, and residual variances. This method performs risk decomposition
        by computing how much each systematic factor and idiosyncratic risk contribute to
        the total portfolio variance.
        The calculation follows the standard risk attribution framework:
        - Portfolio variance = w'Bfactor_covB'w + w'Dresidual_varw
        - Factor risk contribution = (factor_exposure * marginal_contribution) / total_variance
        - Marginal contribution = factor_covariance @ factor_exposure
        Args:
            weights: Portfolio weights DataFrame with dates as index and assets as columns.
                    Each row represents portfolio weights at a specific date, with columns
                    corresponding to individual assets/securities.
            factor_var_name: Name for the factor risk column in output. Defaults to 'Systematic'.
            idiosyncratic_var_name: Name for the idiosyncratic risk column in output. Defaults to 'Idiosyncratic'.
        Returns:
            A tuple containing five elements:
                factor_rcs_ratios: DataFrame of normalized risk contribution ratios where each
                                 row sums to 1.0, including both systematic factors and
                                 idiosyncratic risk.
                factor_risk_contrib_idio: DataFrame of absolute factor risk contributions
                                        normalized by total portfolio variance (systematic + idiosyncratic).
                factor_risk_contrib: DataFrame of factor risk contributions normalized by
                                   systematic variance only (excludes idiosyncratic risk).
                portfolio_var: DataFrame containing systematic (factor-based) variance and idiosyncratic variance for each date.
        Raises:
            ValueError: If x_covars (factor covariance matrices) is None.
            ValueError: If residual_vars (asset-specific residual variances) is None.
        """
        if self.x_covars is None:
            raise ValueError(f"self.x_covars must be provided")
        if self.residual_vars is None:
            raise ValueError(f"self.residual_vars must be provided")
        # Initialize dictionaries to store results for each date
        factor_risk_contrib = {}  # Factor contributions normalized by systematic variance
        factor_risk_contrib_idio = {}  # Factor contributions normalized by total variance
        portfolio_factor_vars = {}  # Systematic variance for each date
        idio_vars = {}  # Idiosyncratic variance for each date
        idio_vars_contrib = {}  # Idiosyncratic risk contribution ratios
        # Iterate through each date in the factor covariance matrix timeline
        for date, factor_covar in self.x_covars.items():
            # Find the most recent portfolio weights available up to the current date
            weight_last_update_date = find_upto_date_from_datetime_index(index=weights.index, date=date)
            last_weight = weights.loc[weight_last_update_date, :]
            # Align factor loadings with portfolio weights (same asset universe)
            last_betas = self.get_loadings_at_date(date=date).loc[:, last_weight.index]
            last_residual_var = self.residual_vars.loc[date, last_weight.index]
            # Calculate portfolio's exposure to each factor: factor_exposure = Beta' * weights
            factor_exposures = last_betas @ last_weight
            # Ensure factor exposures align with covariance matrix factors, fill missing with 0
            factor_exposures = factor_exposures.reindex(index=factor_covar.index).fillna(0.0)
            # Calculate systematic portfolio variance: w'BFB'w where F is factor covariance
            portfolio_var_factors = factor_exposures.T @ (factor_covar @ factor_exposures)
            # Calculate idiosyncratic portfolio variance: w'Dw where D is diagonal residual variance matrix
            portfolio_var_idio = last_weight.T @ np.diag(last_residual_var) @ last_weight
            # Calculate total portfolio variance (systematic + idiosyncratic)
            portfolio_total_var = portfolio_var_factors + portfolio_var_idio
            # Calculate marginal contribution of each factor: dVar/dExposure = F * exposure
            marginal_factor_contrib = factor_covar @ factor_exposures
            # Calculate each factor's contribution to variance: exposure * marginal_contribution
            factor_vars = np.multiply(factor_exposures, marginal_factor_contrib)
            # Normalize factor contributions by systematic variance only
            factor_risk_contrib[date] = factor_vars / portfolio_var_factors if portfolio_var_factors > 0.0 \
                else pd.Series(np.nan, index=factor_covar.index)
            portfolio_factor_vars[date] = portfolio_var_factors
            idio_vars[date] = portfolio_var_idio
            # Calculate idiosyncratic risk contribution ratio
            factor_risk_contrib_idio[date] = factor_vars / portfolio_total_var if portfolio_total_var > 0.0 \
                else pd.Series(np.nan, index=factor_covar.index)
            idio_vars_contrib[date] = portfolio_var_idio / portfolio_total_var if portfolio_total_var > 0.0 else np.nan
        # Convert dictionaries to DataFrames with dates as index
        factor_risk_contrib = pd.DataFrame.from_dict(factor_risk_contrib, orient='index')
        factor_risk_contrib_idio = pd.DataFrame.from_dict(factor_risk_contrib_idio, orient='index')
        # Convert variance dictionaries to Series
        portfolio_factor_vars = pd.Series(portfolio_factor_vars, name=factor_var_name)
        idio_vars = pd.Series(idio_vars, name=idiosyncratic_var_name)
        portfolio_var = pd.concat([portfolio_factor_vars, idio_vars], axis=1)
        idio_vars_contrib = pd.Series(idio_vars_contrib)
        # Add idiosyncratic risk contribution as a separate column
        factor_risk_contrib_idio[idiosyncratic_var_name] = idio_vars_contrib.to_numpy()
        # Normalize all risk contributions to sum to 1.0 across factors for each date
        factor_rcs_ratios = factor_risk_contrib_idio.divide(np.nansum(factor_risk_contrib_idio, axis=1, keepdims=True)).fillna(0.0)
        # Return all computed metrics as a tuple
        return factor_rcs_ratios, factor_risk_contrib_idio, factor_risk_contrib, portfolio_var

    def compute_active_factor_risk(self,
                                   portfolio_weights: pd.DataFrame,
                                   benchmark_weights: pd.DataFrame
                                   ) -> Dict[str, pd.DataFrame]:
        """
        qqq
        """
        if self.x_covars is None:
            raise ValueError(f"self.x_covars must be provided")
        if self.residual_vars is None:
            raise ValueError(f"self.residual_vars must be provided")

        # find joint index
        joint_assets = merge_lists_unique(list1=portfolio_weights.columns.to_list(),
                                          list2=benchmark_weights.columns.to_list())
        portfolio_weights = portfolio_weights.reindex(columns=joint_assets).fillna(0.0)
        benchmark_weights = benchmark_weights.reindex(columns=joint_assets).fillna(0.0)
        residual_vars = self.residual_vars.reindex(columns=joint_assets).fillna(0.0)

        # Initialize dictionaries to store results for each date
        portfolio_exposures_ts = {}
        benchmark_exposures_ts = {}
        active_exposures_ts = {}
        factor_marginal_risks_ts = {}
        factor_risk_contributions_ts = {}
        # Iterate through each date in the factor covariance matrix timeline
        for date, factor_covar in self.x_covars.items():
            # Find the most recent portfolio weights available up to the current date
            weight_last_update_date = find_upto_date_from_datetime_index(index=portfolio_weights.index, date=date)

            if weight_last_update_date in portfolio_weights.index and weight_last_update_date in benchmark_weights.index:
                portfolio_weights_t = portfolio_weights.loc[weight_last_update_date, :]
                benchmark_weights_t = benchmark_weights.loc[weight_last_update_date, :]
                asset_betas_t = self.get_loadings_at_date(date=date).reindex(columns=joint_assets).fillna(0.0)
                idiosyncratic_var_t = residual_vars.loc[date, :]

                """
                # todo: marginal risk by position and groupped risks
                marginal_risk, systematic_marginal, idiosyncratic_marginal = calculate_marginal_active_risk(
                    portfolio_weights=portfolio_weights_t.to_numpy(),
                    benchmark_weights=benchmark_weights_t.to_numpy(),
                    asset_betas=asset_betas_t.to_numpy(),
                    factor_covar=factor_covar.to_numpy(),
                    idiosyncratic_var=idiosyncratic_var_t.to_numpy())
                """
                # Calculate factor exposures
                portfolio_exposures = asset_betas_t @ portfolio_weights_t
                benchmark_exposures = asset_betas_t @ benchmark_weights_t
                active_exposures = portfolio_exposures - benchmark_exposures
                factor_marginal_risks = 2.0 * factor_covar @ active_exposures
                factor_risk_contributions = factor_marginal_risks * active_exposures

                portfolio_exposures_ts[date] = portfolio_exposures
                benchmark_exposures_ts[date] = benchmark_exposures
                active_exposures_ts[date] = active_exposures
                factor_marginal_risks_ts[date] = factor_marginal_risks
                factor_risk_contributions_ts[date] = factor_risk_contributions

        # Convert dictionaries to DataFrames with dates as index
        portfolio_exposures = pd.DataFrame.from_dict(portfolio_exposures_ts, orient='index')
        benchmark_exposures = pd.DataFrame.from_dict(benchmark_exposures_ts, orient='index')
        active_exposures = pd.DataFrame.from_dict(active_exposures_ts, orient='index')
        factor_marginal_risks = pd.DataFrame.from_dict(factor_marginal_risks_ts, orient='index')
        factor_risk_contributions = pd.DataFrame.from_dict(factor_risk_contributions_ts, orient='index')
        factor_risk_contributions_rc = factor_risk_contributions.divide(np.nansum(factor_risk_contributions, axis=1, keepdims=True)).fillna(0.0)

        out_dict = dict(portfolio_exposures=portfolio_exposures,
                        benchmark_exposures=benchmark_exposures,
                        active_exposures=active_exposures,
                        factor_marginal_risks=factor_marginal_risks,
                        factor_risk_contributions=factor_risk_contributions,
                        factor_risk_contributions_rc=factor_risk_contributions_rc)

        return out_dict

    def plot_factor_loadings(self,
                             factor: str,
                             var_format: str = '{:,.2f}',
                             time_period: TimePeriod = None,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        """Plot factor loadings time series."""
        df = self.loadings[factor]
        if time_period is not None:
            df = time_period.locate(df)
        pts.plot_time_series(df=df,
                             var_format=var_format,
                             ax=ax, **kwargs)


def compute_benchmarks_beta_attribution_from_prices(portfolio_nav: pd.Series,
                                                    benchmark_prices: pd.DataFrame,
                                                    portfolio_benchmark_betas: pd.DataFrame,
                                                    residual_name: str = 'Alpha',
                                                    time_period: TimePeriod = None
                                                    ) -> pd.DataFrame:
    """Compute benchmark attribution from price data using pre-calculated betas."""
    benchmark_prices = benchmark_prices.reindex(index=portfolio_benchmark_betas.index, method='ffill')
    portfolio_nav = portfolio_nav.reindex(index=portfolio_benchmark_betas.index, method='ffill')
    x = ret.to_returns(prices=benchmark_prices, freq=None)
    x_attribution = (portfolio_benchmark_betas.shift(1)).multiply(x)
    total_attrib = x_attribution.sum(axis=1)
    total = portfolio_nav.pct_change()
    residual = np.subtract(total, total_attrib)
    joint_attrib = pd.concat([x_attribution, residual.rename(residual_name)], axis=1)
    if time_period is not None:
        joint_attrib = time_period.locate(joint_attrib)
    return joint_attrib


def compute_benchmarks_beta_attribution_from_returns(portfolio_returns: pd.Series,
                                                     benchmark_returns: pd.DataFrame,
                                                     portfolio_benchmark_betas: pd.DataFrame,
                                                     residual_name: str = 'Alpha',
                                                     time_period: TimePeriod = None,
                                                     total_name: Optional[str] = None
                                                     ) -> pd.DataFrame:
    """Compute benchmark attribution from return data using pre-calculated betas."""
    if isinstance(benchmark_returns, pd.Series):
        benchmark_returns = benchmark_returns.to_frame()
    # to be replaced with qis
    benchmark_returns = benchmark_returns.reindex(index=portfolio_returns.index)
    x_attribution = (portfolio_benchmark_betas.shift(1)).multiply(benchmark_returns)
    total_attrib = x_attribution.sum(axis=1)
    residual = np.subtract(portfolio_returns, total_attrib)
    joint_attrib = pd.concat([x_attribution, residual.rename(residual_name)], axis=1)
    if total_name is not None:
        joint_attrib = pd.concat([portfolio_returns.rename(total_name), joint_attrib], axis=1)
    if time_period is not None:
        joint_attrib = time_period.locate(joint_attrib)
    joint_attrib.iloc[0, :] = 0.0
    return joint_attrib



"""
implementation for infinite lower and upper bounds for pd.cut with wrapper
https://stackoverflow.com/questions/30127427/pandas-cut-with-infinite-upper-lower-bounds
"""
import numpy as np
import pandas as pd
from typing import Optional, Union, List, Tuple
from pandas import Categorical
from enum import Enum


def x_bins_cut(a: np.ndarray,
               bins: np.ndarray,
               lower_infinite: bool = True,
               upper_infinite: bool = True,
               xvar_format: str = '{:.2f}',
               is_value_labels: bool = True,
               value_label: str = 'regime',
               bucket_prefix: str = None,
               **kwargs
               ) -> Tuple[Categorical, List[str]]:
    r"""Wrapper around pandas cut() to create infinite lower/upper bounds with proper labeling.

    Takes all the same arguments as pandas cut(), plus two more.

    Args :
        lower_infinite (bool, optional) : set whether the lower bound is infinite
            Default is True. If true, and your first bin element is something like 20, the
            first bin label will be '<= 20' (depending on other cut() parameters)
        upper_infinite (bool, optional) : set whether the upper bound is infinite
            Default is True. If true, and your last bin element is something like 20, the
            first bin label will be '> 20' (depending on other cut() parameters)
        **kwargs : any standard pandas cut() labeled parameters

    Returns :
        out : same as pandas cut() return value
        bins : same as pandas cut() return value
    """

    # Quick passthru if no infinite bounds
    if not lower_infinite and not upper_infinite:
        return pd.cut(a, bins, **kwargs)

    # Setup
    num_labels = len(bins) - 1
    include_lowest = kwargs.get("include_lowest", False)
    right = kwargs.get("right", True)

    # Prepend/Append infinities where indiciated
    bins_final = bins.copy()
    if upper_infinite:
        bins_final = np.append(bins_final, float("inf"))
        num_labels += 1
    if lower_infinite:
        bins_final = np.append(float("-inf"), bins_final)
        num_labels += 1

    # Decide all boundary symbols based on traditional cut() parameters
    symbol_lower = "<=" if include_lowest and right else "<"
    left_bracket = u"=" + "(" if right else "["
    right_bracket = "]" if right else ")"
    symbol_upper = ">" if right else ">="

    # Inner function reused in multiple clauses for labeling
    def make_label(i: int, lb: str = left_bracket, rb: str = right_bracket):
        if is_value_labels:
            label = f"{lb}{xvar_format.format(bins_final[i])}, {xvar_format.format(bins_final[i+1])}{rb}"
        else:
            label = f"{value_label}-{i+1}"
        return label

    # Create custom labels
    labels = []
    for i in range(0, num_labels):
        if i == 0 and is_value_labels:
            if lower_infinite:
                new_label = f"{symbol_lower}{xvar_format.format(bins_final[i+1])}"
            elif include_lowest:
                new_label = make_label(i, lb="[")
            else:
                new_label = make_label(i)
        elif upper_infinite and i == (num_labels - 1) and is_value_labels:
            new_label = f"{symbol_upper}{xvar_format.format(bins_final[i])}"
        else:
            new_label = make_label(i)
        if bucket_prefix is not None:
            new_label = f"{bucket_prefix}{new_label}"
        labels.append(new_label)
    try:
        out = pd.cut(a, bins_final, labels=labels)
    except ValueError:   # labels must be unique if ordered=True
        labels = [f"bin_{n+1}" for n, _ in enumerate(bins_final[1:])]
        out = pd.cut(a, bins_final, labels=labels)
    return out, labels


def add_classification(df: pd.DataFrame,
                       class_var_col: str,
                       bins: np.ndarray = np.array([-3.0, -1.5, 0.0, 1.5, 3.0]),
                       hue_name: str = 'hue',
                       xvar_format: str = '{:.2f}',
                       is_value_labels: bool = True,
                       is_to_str: bool = True,
                       bucket_prefix: str = None,
                       **kwargs
                       ) -> Tuple[pd.DataFrame, List[str]]:

    df = df.copy()
    df = df.sort_values(by=class_var_col)
    class_var, labels = x_bins_cut(a=df[class_var_col],
                                   bins=bins,
                                   is_value_labels=is_value_labels,
                                   xvar_format=xvar_format,
                                   bucket_prefix=bucket_prefix,
                                   **kwargs)
    df[hue_name] = class_var
    if is_to_str:
        df[hue_name] = df[hue_name].astype(str)
    return df, labels


def add_quantile_classification(df: pd.DataFrame,
                                x_column: str,
                                hue_name: str = 'hue',
                                num_buckets: Optional[Union[np.ndarray, int]] = None,
                                bins: Optional[np.ndarray] = np.array([-3.0, -1.5, 0.0, 1.5, 3.0]),
                                xvar_format: str = '{:.2f}',
                                is_value_labels: bool = True,
                                bucket_prefix: str = None,
                                **kwargs
                                ) -> Tuple[pd.DataFrame, List[str]]:
    if num_buckets is not None:  # create bins
        if isinstance(num_buckets, int):
            bins = np.nanquantile(df[x_column], q=[(1.0 / num_buckets) * (n + 1) for n in np.arange(num_buckets - 1)])
        else:
            bins = np.nanquantile(df[x_column], q=num_buckets)
    df, labels = add_classification(df=df,
                                    class_var_col=x_column,
                                    bins=bins,
                                    hue_name=hue_name,
                                    is_value_labels=is_value_labels,
                                    xvar_format=xvar_format,
                                    bucket_prefix=bucket_prefix,
                                    **kwargs)
    return df, labels


def sort_index_by_hue(df: pd.DataFrame, hue_order: List[str]) -> pd.DataFrame:
    sort_column = 'sort_column'
    name_sort = {key: idx for idx, key in enumerate(hue_order)}
    df[sort_column] = df.index.map(name_sort)
    df = df.sort_values(by=sort_column).drop(sort_column, axis=1)
    return df


def add_hue_fixed_years(df: pd.DataFrame,
                        hue: str,
                        fixed_years: List[int] = (2001, 2006, 2010, 2018, 2021, 2022)
                        ) -> pd.DataFrame:
    """
    map sequence of years into hue = (fixed_years[i-1], fixed_years[i]]
    """
    pd_idx = pd.IntervalIndex.from_breaks(fixed_years, closed='right')
    mapper = {idx: f"{p.left+1}-{p.right}" for idx, p in enumerate(pd_idx)}
    hue_vals = pd.Series([pd_idx.get_loc(x.year) for x in df.index]).map(mapper)
    df[hue] = hue_vals.to_list()
    return df


def add_hue_years(df: pd.DataFrame, hue: str) -> pd.DataFrame:
    """
    add hue in years
    """
    df[hue] = [x.year for x in df.index]
    return df


class LocalTests(Enum):
    CUT = 1
    CLASS = 2


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    np.random.seed(2)  # freeze seed

    if local_test == LocalTests.CUT:

        n = 1000000
        x = np.random.normal(0.0, 1.0, n)
        bins = np.array([-3.0, -1.5, 0.0, 1.5, 3.0])

        pd1 = pd.cut(x, bins)
        print('pd.cut')
        print(pd1)

        pd2, labels = x_bins_cut(x, bins)
        print('x_bins_cut')
        print(pd2)

    elif local_test == LocalTests.CLASS:

        n = 10000
        x = np.random.normal(0.0, 1.0, n)
        eps = np.random.normal(0.0, 1.0, n)
        y = x + eps
        bins = np.array([-3.0, -1.5, 0.0, 1.5, 3.0])
        df = pd.concat([pd.Series(x, name='x'), pd.Series(y, name='y')], axis=1)

        df1, labels = add_classification(df=df, class_var_col='x', bins=bins)
        print(df1)


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.CUT)



"""
Brinson Performance Attribution Analysis

This module implements the Brinson performance attribution model for decomposing
active returns between portfolios and benchmarks into allocation and selection effects.

References:
    Brinson, G.P., Hood, L.R., & Beebower, G.L. (1986). Determinants of portfolio performance.
    https://en.wikipedia.org/wiki/Performance_attribution
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from typing import Tuple, List, Optional
from enum import Enum

import qis as qis
from qis.utils.df_groups import agg_df_by_groups
from qis.utils.struct_ops import merge_lists_unique
import qis.utils.df_agg as dfa
import qis.plots.time_series as pts
from qis.plots.table import plot_df_table


def compute_brinson_attribution_table(
    benchmark_pnl: pd.DataFrame,
    strategy_pnl: pd.DataFrame,
    strategy_weights: pd.DataFrame,
    benchmark_weights: pd.DataFrame,
    asset_class_data: pd.Series,
    group_order: List[str] = None,
    total_column: str = 'Total Sum',
    is_exclude_interaction_term: bool = True,
    strategy_name: str = 'Strategy',
    benchmark_name: str = 'Benchmark'
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Compute Brinson performance attribution decomposition.

    Decomposes active returns into allocation and selection effects using the Brinson model:
    - Allocation Effect: (w_p - w_b) × r_b (overweight/underweight better/worse performers)
    - Selection Effect: w_b × (r_p - r_b) (security selection within asset classes)
    - Interaction Effect: (w_p - w_b) × (r_p - r_b) (combined allocation & selection)

    Where:
        w_p, w_b = portfolio and benchmark weights
        r_p, r_b = portfolio and benchmark returns

    Args:
        benchmark_pnl: Benchmark P&L by asset over time. Shape: (T, N_assets).
        strategy_pnl: Strategy P&L by asset over time. Shape: (T, N_assets).
        strategy_weights: Strategy weights by asset over time. Shape: (T, N_assets).
        benchmark_weights: Benchmark weights by asset over time. Shape: (T, N_assets).
        asset_class_data: Mapping of assets to asset classes. Index: assets, Values: classes.
        group_order: Optional ordering for asset classes in output tables.
        total_column: Name for the total/summary column in outputs.
        is_exclude_interaction_term: If True, allocates interaction effect to selection.
        strategy_name: Display name for strategy in output tables.
        benchmark_name: Display name for benchmark in output tables.

    Returns:
        Tuple containing:
            - totals_table: Summary statistics by asset class (weights, returns, attribution)
            - active_total: Time series of total allocation and selection effects
            - grouped_allocation_return: Allocation effect by asset class over time
            - grouped_selection_return: Selection effect by asset class over time
            - grouped_interaction_return: Interaction effect by asset class over time

    Raises:
        ValueError: If input DataFrames have mismatched dimensions or invalid data.

    Example:
        >>> benchmark_pnl = pd.DataFrame(...)  # Daily benchmark returns by asset
        >>> strategy_pnl = pd.DataFrame(...)   # Daily strategy returns by asset
        >>> strategy_weights = pd.DataFrame(...)  # Daily strategy weights by asset
        >>> benchmark_weights = pd.DataFrame(...) # Daily benchmark weights by asset
        >>> asset_classes = pd.Series({'AAPL': 'Equities', 'TLT': 'Bonds', ...})
        >>>
        >>> results = compute_brinson_attribution_table(
        ...     benchmark_pnl, strategy_pnl, strategy_weights,
        ...     benchmark_weights, asset_classes
        ... )
        >>> totals, active_ts, allocation, selection, interaction = results
    """
    # Data alignment and preprocessing
    # Create unified asset universe by merging strategy and benchmark assets
    joint_assets = merge_lists_unique(
        benchmark_pnl.columns.to_list(),
        strategy_pnl.columns.to_list()
    )

    # Ensure all datasets have consistent asset coverage
    # Fill missing assets with appropriate defaults (0 for weights/PnL, 'Unclassified' for classes)
    asset_class_data = asset_class_data.reindex(index=joint_assets).fillna('Unclassified')
    strategy_pnl = strategy_pnl.reindex(columns=joint_assets).fillna(0.0)
    benchmark_pnl = benchmark_pnl.reindex(columns=joint_assets).fillna(0.0)
    strategy_weights = strategy_weights.reindex(columns=joint_assets).fillna(0.0)
    benchmark_weights = benchmark_weights.reindex(columns=joint_assets).fillna(0.0)

    # Asset class aggregation
    # Group individual assets into asset classes and sum P&L within each class
    grouped_strategy_pnl = agg_df_by_groups(
        df=strategy_pnl,
        group_data=asset_class_data,
        group_order=group_order,
        agg_func=dfa.nansum
    )

    grouped_benchmark_pnl = agg_df_by_groups(
        df=benchmark_pnl,
        group_data=asset_class_data,
        group_order=group_order,
        agg_func=dfa.nansum
    )

    # Aggregate weights by asset class (sum individual asset weights within class)
    grouped_strategy_weights = agg_df_by_groups(
        df=strategy_weights,
        group_data=asset_class_data,
        group_order=group_order,
        agg_func=dfa.nansum
    )

    grouped_benchmark_weights = agg_df_by_groups(
        df=benchmark_weights,
        group_data=asset_class_data,
        group_order=group_order,
        agg_func=dfa.nansum
    )

    # Core Brinson attribution calculations
    # Active return computation (two equivalent approaches)
    active_return = strategy_pnl - benchmark_pnl

    # Flag to choose between direct aggregation vs grouped calculation
    # Both should yield identical results - using grouped for consistency
    is_revaluate_active_return = False
    if is_revaluate_active_return:
        # Approach 1: Aggregate individual asset active returns
        grouped_active_return = agg_df_by_groups(
            df=active_return,
            group_data=asset_class_data,
            group_order=group_order,
            agg_func=dfa.nansum
        )
    else:
        # Approach 2: Calculate active return from grouped P&L (preferred)
        grouped_active_return = grouped_strategy_pnl - grouped_benchmark_pnl

    # Brinson allocation effect: (w_strategy - w_benchmark) × r_benchmark
    # Measures value added/lost from asset allocation decisions
    grouped_allocation_return = (
        (grouped_strategy_weights - grouped_benchmark_weights) * grouped_benchmark_pnl
    )

    # Brinson selection effect: w_benchmark × (r_strategy - r_benchmark)
    # Measures value added/lost from security selection within asset classes
    grouped_selection_return = (
        grouped_benchmark_weights * (grouped_strategy_pnl - grouped_benchmark_pnl)
    )

    # Interaction effect: Residual not captured by allocation + selection
    # Represents the interaction between allocation and selection decisions
    grouped_interaction_return = (
        grouped_active_return - (grouped_allocation_return + grouped_selection_return)
    )

    # Interaction term handling
    if is_exclude_interaction_term:
        # Common practice: Allocate interaction effect to selection component
        # Alternative approaches: split 50/50 between allocation and selection
        grouped_selection_return = grouped_selection_return + grouped_interaction_return
        grouped_interaction_return = 0.0 * grouped_interaction_return

    # Summary statistics table creation
    # Create comprehensive summary with time-aggregated statistics by asset class
    totals_table = pd.DataFrame(index=grouped_strategy_pnl.columns.to_list() + [total_column])

    # Average weights over the analysis period
    totals_table[f"{strategy_name}\nWeight Ave"] = dfa.agg_data_by_axis(
        df=grouped_strategy_weights,
        agg_func=np.nanmean,
        total_column=total_column
    )
    totals_table[f"{benchmark_name}\nWeight Ave"] = dfa.agg_data_by_axis(
        df=grouped_benchmark_weights,
        agg_func=np.nanmean,
        total_column=total_column
    )

    # Cumulative returns over the analysis period
    totals_table[f"{strategy_name}\nReturn Sum"] = dfa.agg_data_by_axis(
        df=grouped_strategy_pnl,
        agg_func=np.nansum,
        total_column=total_column
    )
    totals_table[f"{benchmark_name}\nReturn Sum"] = dfa.agg_data_by_axis(
        df=grouped_benchmark_pnl,
        agg_func=np.nansum,
        total_column=total_column
    )

    # Attribution effects summed over the analysis period
    totals_table['Asset\nAllocation'] = dfa.agg_data_by_axis(
        df=grouped_allocation_return,
        agg_func=np.nansum,
        total_column=total_column
    )
    totals_table['Instrument\nSelection'] = dfa.agg_data_by_axis(
        df=grouped_selection_return,
        agg_func=np.nansum,
        total_column=total_column
    )

    # Include interaction term in summary if not excluded
    if not is_exclude_interaction_term:
        totals_table['Interaction'] = dfa.agg_data_by_axis(
            df=grouped_interaction_return,
            agg_func=np.nansum,
            total_column=total_column
        )

    # Total active return for verification (should equal allocation + selection + interaction)
    totals_table['Total\nActive'] = dfa.agg_data_by_axis(
        df=grouped_active_return,
        agg_func=np.nansum,
        total_column=total_column
    )

    # Time series preparation
    # Add portfolio-level totals (sum across all asset classes) for time series analysis
    grouped_allocation_return[total_column] = np.sum(grouped_allocation_return, axis=1)
    grouped_selection_return[total_column] = np.sum(grouped_selection_return, axis=1)
    grouped_interaction_return[total_column] = np.sum(grouped_interaction_return, axis=1)

    # Create time series of total attribution effects for plotting
    allocation_total = grouped_allocation_return[total_column].to_frame(name='Allocation Total')
    selection_total = grouped_selection_return[total_column].to_frame(name='Selection Total')
    active_total = pd.concat([allocation_total, selection_total], axis=1)

    if not is_exclude_interaction_term:
        interaction_total = grouped_interaction_return[total_column].to_frame(name='Interaction Total')
        active_total = pd.concat([active_total, interaction_total], axis=1)

    return (totals_table, active_total, grouped_allocation_return,
            grouped_selection_return, grouped_interaction_return)


def plot_brinson_attribution_table(
    totals_table: pd.DataFrame,
    active_total: pd.DataFrame,
    grouped_allocation_return: pd.DataFrame,
    grouped_selection_return: pd.DataFrame,
    grouped_interaction_return: pd.DataFrame,
    var_format: str = '{:.0%}',
    total_column: str = 'Total Sum',
    is_exclude_interaction_term: bool = True,
    axs: List[plt.Subplot] = (None, None, None, None, None),
    **kwargs
) -> Tuple[plt.Figure, plt.Figure, plt.Figure, plt.Figure, plt.Figure]:
    """Create comprehensive visualization of Brinson attribution results.

    Generates a multi-panel visualization including:
    1. Summary table with attribution statistics
    2. Time series of cumulative total attribution effects
    3. Time series of cumulative allocation effects by asset class
    4. Time series of cumulative selection effects by asset class
    5. Time series of cumulative interaction effects by asset class

    Args:
        totals_table: Summary statistics table from compute_brinson_attribution_table.
        active_total: Time series of total attribution effects.
        grouped_allocation_return: Allocation effects by asset class over time.
        grouped_selection_return: Selection effects by asset class over time.
        grouped_interaction_return: Interaction effects by asset class over time.
        var_format: Format string for displaying numeric values (default: percentage).
        total_column: Name of the total column for portfolio-level aggregation.
        is_exclude_interaction_term: Whether interaction terms are excluded from analysis.
        axs: Optional list of matplotlib axes for plotting (if None, creates new figures).
        **kwargs: Additional arguments passed to plotting functions.

    Returns:
        Tuple of matplotlib figures: (table_fig, active_fig, allocation_fig,
                                    selection_fig, interaction_fig)

    Example:
        >>> # After running compute_brinson_attribution_table
        >>> figs = plot_brinson_attribution_table(
        ...     totals_table, active_total, allocation_return,
        ...     selection_return, interaction_return
        ... )
        >>> table_fig, active_fig, alloc_fig, select_fig, interact_fig = figs
        >>> plt.show()
    """
    # Generate formatted summary table
    fig_table = plot_brinson_totals_table(
        totals_table=totals_table,
        var_format=var_format,
        ax=axs[0],
        **kwargs
    )

    # Plot cumulative total attribution effects over time
    # Convert to cumulative for better visualization of performance evolution
    active_total_cumsum = active_total.cumsum(axis=0)
    fig_active_total = pts.plot_time_series(
        df=active_total_cumsum,
        var_format='{:.0%}',
        title='Cumulative Active Attribution Effects',
        legend_stats=qis.LegendStats.LAST_NONNAN,
        ax=axs[1],
        **kwargs
    )

    # Plot cumulative allocation effects by asset class
    cum_allocation_return = grouped_allocation_return.cumsum(axis=0)
    fig_ts_alloc = pts.plot_time_series(
        df=cum_allocation_return,
        var_format='{:.0%}',
        title='Cumulative Asset Class Allocation Effects',
        legend_stats=qis.LegendStats.LAST_NONNAN,
        ax=axs[2],
        **kwargs
    )

    # Plot cumulative selection effects by asset class
    cum_selection_return = grouped_selection_return.cumsum(axis=0)
    fig_ts_sel = pts.plot_time_series(
        df=cum_selection_return,
        var_format='{:.0%}',
        title='Cumulative Asset Class Selection Effects',
        legend_stats=qis.LegendStats.LAST_NONNAN,
        ax=axs[3],
        **kwargs
    )

    # Plot cumulative interaction effects by asset class (with trend line)
    cum_interaction_return = grouped_interaction_return.cumsum(axis=0)
    fig_ts_inter = pts.plot_time_series(
        df=cum_interaction_return,
        trend_line=pts.TrendLine.TREND_LINE,
        var_format='{:.0%}',
        title='Cumulative Asset Class Interaction Effects',
        ax=axs[4],
        **kwargs
    )

    return fig_table, fig_active_total, fig_ts_alloc, fig_ts_sel, fig_ts_inter


def plot_brinson_totals_table(
    totals_table: pd.DataFrame,
    var_format: str = '{:.0%}',
    ax: Optional[plt.Subplot] = None,
    **kwargs
) -> Optional[plt.Figure]:
    """Create formatted table visualization of Brinson attribution summary.

    Generates a professional-looking table with:
    - Color-coded sections for different data types
    - Edge lines separating logical groups
    - Formatted numeric values
    - Highlighted totals row

    Args:
        totals_table: Summary statistics DataFrame with attribution results.
        var_format: String format for numeric values (default: percentage format).
        ax: Optional matplotlib axis for plotting. If None, creates new figure.
        **kwargs: Additional arguments passed to plot_df_table function.

    Returns:
        matplotlib Figure object containing the formatted table.

    Example:
        >>> fig = plot_brinson_totals_table(totals_table, var_format='{:.2%}')
        >>> plt.show()
    """
    # Define visual formatting for the table
    # Highlight the totals row at the bottom
    special_rows_colors = [(len(totals_table.index), 'steelblue')]
    rows_edge_lines = [len(totals_table.index) - 1]  # Separator line before totals

    # Color-code different column groups
    special_columns_colors = [
        (0, 'lightblue'),  # First column (asset class names)
        (len(totals_table.columns), 'steelblue')  # Last column
    ]

    # Add vertical separator lines between logical column groups
    columns_edge_lines = [
        (1, 'black'),  # After asset class names
        (3, 'black'),  # After weight columns
        (5, 'black'),  # After return columns
        (8, 'black')   # After attribution columns
    ]

    # Format all numeric values according to specified format
    totals_table_formatted = qis.df_to_str(df=totals_table, var_format=var_format)

    # Generate the formatted table plot
    fig_table = plot_df_table(
        df=totals_table_formatted,
        column_width=2.0,
        first_column_width=2.0,
        special_rows_colors=special_rows_colors,
        rows_edge_lines=rows_edge_lines,
        special_columns_colors=special_columns_colors,
        columns_edge_lines=columns_edge_lines,
        ax=ax,
        **kwargs
    )
    return fig_table


class LocalTests(Enum):
    """Enumeration of available local test scenarios."""
    ATTRIBUTION = 1


def run_local_test(local_test: LocalTests) -> None:
    """Execute local integration tests for development and validation.

    Provides comprehensive test scenarios with realistic data to validate
    the Brinson attribution implementation. These tests include edge cases
    such as assets present in only one portfolio (strategy vs benchmark).

    Args:
        local_test: Test scenario to execute (currently only ATTRIBUTION available).

    Example:
        >>> run_local_test(LocalTests.ATTRIBUTION)
        # Executes attribution test with mixed asset universe and displays results
    """
    if local_test == LocalTests.ATTRIBUTION:
        # Test data setup with realistic scenarios
        dates = pd.date_range('2024-01-01', periods=3, freq='D')

        # Define asset class ordering for consistent output formatting
        group_order = ['Equities', 'Bonds', 'ALT_IN_STRATEGY', 'COMMODITIES_IN_BENCHMARK']

        # Asset class mapping including assets unique to each portfolio
        asset_classes = pd.Series({
            'STOCK_A': 'Equities',
            'STOCK_B': 'Equities',
            'BOND_A': 'Bonds',
            'BOND_B': 'Bonds',
            'ALT_IN_STRATEGY': 'ALT_IN_STRATEGY',  # Only in strategy portfolio
            'COMMODITIES_IN_BENCHMARK': 'COMMODITIES_IN_BENCHMARK'  # Only in benchmark
        })

        # Strategy P&L data (equities outperform, alternatives included)
        strategy_pnl = pd.DataFrame({
            'STOCK_A': [0.02, 0.015, 0.01],     # Strong equity performance
            'STOCK_B': [0.018, 0.012, 0.008],   # Strong equity performance
            'BOND_A': [0.005, 0.003, 0.002],    # Modest bond performance
            'BOND_B': [0.004, 0.002, 0.001],    # Modest bond performance
            'ALT_IN_STRATEGY': [0.04, 0.02, 0.01]  # High alternative returns
        }, index=dates)

        # Benchmark P&L data (similar patterns, commodities included)
        benchmark_pnl = pd.DataFrame({
            'STOCK_A': [0.018, 0.013, 0.009],
            'STOCK_B': [0.016, 0.010, 0.007],
            'BOND_A': [0.004, 0.002, 0.001],
            'BOND_B': [0.003, 0.001, 0.0005],
            'COMMODITIES_IN_BENCHMARK': [0.03, 0.01, 0.005]  # Commodity exposure
        }, index=dates)

        # Strategy weights (overweight equities, include alternatives)
        strategy_weights = pd.DataFrame({
            'STOCK_A': [0.35, 0.35, 0.35],                    # Overweight equities
            'STOCK_B': [0.25, 0.25, 0.25],                    # Overweight equities
            'BOND_A': [0.25, 0.25, 0.25],                     # Standard bond allocation
            'BOND_B': [0.15 / 2.0, 0.15 / 2.0, 0.15 / 2.0],   # Split remaining weight
            'ALT_IN_STRATEGY': [0.15 / 2.0, 0.15 / 2.0, 0.15 / 2.0]  # Alternative allocation
        }, index=dates)

        # Benchmark weights (balanced allocation, include commodities)
        benchmark_weights = pd.DataFrame({
            'STOCK_A': [0.25, 0.25, 0.25],                    # Equal equity weight
            'STOCK_B': [0.25, 0.25, 0.25],                    # Equal equity weight
            'BOND_A': [0.25, 0.25, 0.25],                     # Equal bond weight
            'BOND_B': [0.25 / 2.0, 0.25 / 2.0, 0.25 / 2.0],   # Split remaining weight
            'COMMODITIES_IN_BENCHMARK': [0.25 / 2.0, 0.25 / 2.0, 0.25 / 2.0]  # Commodity allocation
        }, index=dates)

        # Execute Brinson attribution analysis
        (totals_table, active_total, grouped_allocation_return,
         grouped_selection_return, grouped_interaction_return) = compute_brinson_attribution_table(
            benchmark_pnl=benchmark_pnl,
            strategy_pnl=strategy_pnl,
            strategy_weights=strategy_weights,
            benchmark_weights=benchmark_weights,
            asset_class_data=asset_classes,
            group_order=group_order
        )

        # Display results
        print("=== BRINSON ATTRIBUTION ANALYSIS RESULTS ===")
        print(totals_table)

        # Generate and display summary table visualization
        plot_brinson_totals_table(totals_table=totals_table)
        plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.ATTRIBUTION)




import pandas as pd
from enum import Enum
from qis.utils.dates import (TimePeriod,
                             shift_dates_by_year,
                             generate_sample_dates,
                             get_weekday,
                             get_sample_dates_idx,
                             generate_dates_schedule,
                             FreqMap,
                             generate_rebalancing_indicators,
                             generate_fixed_maturity_rolls,
                             find_upto_date_from_datetime_index,
                             create_rebalancing_indicators_from_freqs)

class LocalTests(Enum):
    DATES = 1
    OFFSETS = 2
    WEEK_DAY = 3
    SAMPLE_DATES_IDX = 4
    PERIOD_WITH_HOLIDAYS = 5
    EOD_FREQ_HOUR = 6
    FREQ_HOUR = 7
    FREQ_REB = 8
    FREQS = 9
    REBALANCING_INDICATORS = 10
    FIXED_MATURITY_ROLLS = 11
    TIMEPERIOD_INDEXER = 12
    FIND_UPTO_DATE = 14
    REBALANCING_FROM_FREQS = 15


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    if local_test == LocalTests.DATES:
        time_period = TimePeriod(start='31Dec2019', end='31Dec2020')
        print("time_period.to_str() = ")
        print(time_period.to_str())
        print("time_period.start_to_str() = ")
        print(time_period.start_to_str())
        print("time_period.end_to_str() = ")
        print(time_period.end_to_str())
        print("time_period.to_pd_datetime_index(freq='ME') = ")
        print(time_period.to_pd_datetime_index(freq='ME'))
        print("time_period.to_pd_datetime_index(freq='ME') = ")
        print(time_period.to_pd_datetime_index(freq='ME', tz='America/New_York'))
        print("time_period.get_period_dates_str(freq='ME', tz='America/New_York') = ")
        print(time_period.to_period_dates_str(freq='ME'))

    elif local_test == LocalTests.OFFSETS:
        date = pd.Timestamp('28Feb2023')
        print(date-pd.offsets.Day(30))
        print(date-pd.offsets.MonthEnd(1))

        print(shift_dates_by_year(date))
        print(date-pd.offsets.Day(365))
        print(date-pd.offsets.BYearBegin(1))

    elif local_test == LocalTests.WEEK_DAY:
        time_period = TimePeriod(start='01Jun2022', end='22Jun2022')
        sample_dates = generate_sample_dates(time_period=time_period, freq='D')
        print(get_weekday(sample_dates.index))

    elif local_test == LocalTests.SAMPLE_DATES_IDX:
        time_period = TimePeriod(start='31Dec2004', end='31Dec2020')
        population_dates = time_period.to_pd_datetime_index(freq='ME')
        sample_dates = time_period.to_pd_datetime_index(freq='QE')
        sample_dates_idx = get_sample_dates_idx(population_dates=population_dates, sample_dates=sample_dates)

        print('population_dates')
        print(population_dates)
        print('sample_dates')
        print(sample_dates)
        print('sample_dates_idx')
        print(sample_dates_idx)

        print('sampled_population_dates')
        print(population_dates[sample_dates_idx])

    elif local_test == LocalTests.PERIOD_WITH_HOLIDAYS:

        time_period = TimePeriod(start='01Jan2020', end='22Jan2020')
        holidays = pd.DatetimeIndex(data=['20200113', '20200120', '20200214', '20200217',
                                          '20200217', '20200219', '20200228', '20200324', '20200410'])

        dates = time_period.to_period_dates_str(freq='B', holidays=holidays)
        print(dates)

    elif local_test == LocalTests.EOD_FREQ_HOUR:
        time_period = TimePeriod(pd.Timestamp('2022-11-10', tz='UTC'), pd.Timestamp('2023-03-03', tz='UTC'))
        rebalancing_times = generate_dates_schedule(time_period=time_period, freq='h')
        print(rebalancing_times)

    elif local_test == LocalTests.FREQ_HOUR:

        # hour frequency with timestamp with no hour
        time_period = TimePeriod(pd.Timestamp('2022-11-10', tz='UTC'), pd.Timestamp('2022-11-16', tz='UTC'))
        rebalancing_times = generate_dates_schedule(time_period=time_period, freq='8H')
        print(rebalancing_times)

        rebalancing_times = generate_dates_schedule(time_period=time_period, freq='D_8H')
        print(rebalancing_times)

        value_time = pd.Timestamp('2023-10-04 08:00:00+00:00').normalize()
        print(value_time)
        time_period = TimePeriod(value_time, value_time)
        rebalancing_times = generate_dates_schedule(time_period=time_period, freq='h')
        print(rebalancing_times)

    elif local_test == LocalTests.FREQ_REB:
        dates_schedule = generate_dates_schedule(time_period=TimePeriod('2023-05-01 08:00:00', '2023-05-30 10:00:00', tz='UTC'),
                                                 freq='2W-FRI', hour_offset=8,
                                                 include_start_date=True,
                                                 include_end_date=True)
        print(dates_schedule)
        dates_schedule = generate_dates_schedule(time_period=TimePeriod('2023-05-01 08:00:00', '2023-05-30 10:00:00', tz='UTC'),
                                                 freq='2W-FRI', hour_offset=8,
                                                 include_start_date=False,
                                                 include_end_date=False)
        print(dates_schedule)

    elif local_test == LocalTests.FREQS:
        freq_map = FreqMap.BQ
        freq_map.print()
        freq = freq_map.to_freq()
        print(freq)
        n_bus_days = freq_map.to_n_bus_days()
        print(n_bus_days)

    elif local_test == LocalTests.REBALANCING_INDICATORS:
        pd_index = pd.date_range(start='31Dec2020', end='31Dec2021', freq='W-MON')
        data = pd.DataFrame(range(len(pd_index)), index=pd_index, columns=['aaa'])
        rebalancing_schedule = generate_rebalancing_indicators(df=data, freq='M-FRI')
        print(rebalancing_schedule)
        print(rebalancing_schedule[rebalancing_schedule == True])

    elif local_test == LocalTests.FIXED_MATURITY_ROLLS:
        time_period = TimePeriod('01Oct2022', '21Feb2023', tz='UTC')
        """
        weekly_rolls = generate_fixed_maturity_rolls(time_period=time_period, freq='h', roll_freq='W-FRI',
                                                     roll_hour=8,
                                                     min_days_to_next_roll=6)
        print(f"weekly_rolls:\n{weekly_rolls}")

        monthly_rolls = generate_fixed_maturity_rolls(time_period=time_period, freq='h', roll_freq='M-FRI',
                                                      roll_hour=8,
                                                      min_days_to_next_roll=28)  # 4 weeks before
        print(f"monthly_rolls:\n{monthly_rolls}")
        """
        quarterly_rolls = generate_fixed_maturity_rolls(time_period=time_period, freq='h', roll_freq='Q-FRI',
                                                        roll_hour=8,
                                                        min_days_to_next_roll=28)  # 4 weeks before
        print(f"quarterly_rolls:\n{quarterly_rolls}")
        quarterly_rolls = generate_fixed_maturity_rolls(time_period=time_period, freq='h', roll_freq='Q-FRI',
                                                        roll_hour=8,
                                                        min_days_to_next_roll=56)  # 8 weeks before
        print(f"quarterly_rolls:\n{quarterly_rolls}")

    elif local_test == LocalTests.TIMEPERIOD_INDEXER:
        time_period = TimePeriod('31Dec2023', '31Dec2024')
        times_m = generate_dates_schedule(time_period=time_period, freq='ME')
        times_q = generate_dates_schedule(time_period=time_period, freq='QE')
        print(times_m)
        print(times_q)
        idx_q_at_m = times_m.get_indexer(target=times_q, method='ffill')
        print(idx_q_at_m)
        idx_m_at_q = times_q.get_indexer(target=times_m, method='ffill')
        print(idx_m_at_q)
        indextimes_m_at_q = [times_q[idx] for idx in idx_m_at_q]
        print(indextimes_m_at_q)

    elif local_test == LocalTests.FIND_UPTO_DATE:
        time_period = TimePeriod('31Dec2023', '31Dec2024')
        times_q = generate_dates_schedule(time_period=time_period, freq='QE')
        this_dates = [pd.Timestamp('31Dec2023'), pd.Timestamp('31Jan2024'), pd.Timestamp('31Mar2024'),
                      pd.Timestamp('31Dec2024'), pd.Timestamp('31Dec2025')]
        for this_date in this_dates:
            matched_date = find_upto_date_from_datetime_index(index=times_q, date=this_date)
            print(f"given={this_date}, matched={matched_date}")

    elif local_test == LocalTests.REBALANCING_FROM_FREQS:
        time_period = TimePeriod('31Dec2023', '31Dec2024')
        rebalancing_indicators = create_rebalancing_indicators_from_freqs(
            rebalancing_freqs=pd.Series(dict(SPY='ME', TLT='QE', LQD='QE')),
            time_period=time_period)
        print(rebalancing_indicators)


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.REBALANCING_FROM_FREQS)
